import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff.parameters import (
    AngleHandler,
    BondHandler,
    ImproperTorsionHandler,
)
from openff.utilities import has_package, skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import requires_openeye
from openff.interchange.constants import kcal_mol_a2, kcal_mol_rad2
from openff.interchange.exceptions import DuplicateMoleculeError
from openff.interchange.models import AngleKey, BondKey, ImproperTorsionKey
from openff.interchange.smirnoff._create import _create_interchange
from openff.interchange.smirnoff._valence import (
    SMIRNOFFAngleCollection,
    SMIRNOFFBondCollection,
    SMIRNOFFConstraintCollection,
    SMIRNOFFImproperTorsionCollection,
    _check_molecule_uniqueness,
)

if has_package("openmm"):
    import openmm.app
    import openmm.unit

from pydantic import ValidationError


class TestSMIRNOFFValenceCollections:
    def test_bond_collection(self, water):
        bond_handler = BondHandler(version=0.3)
        bond_handler.fractional_bondorder_method = "AM1-Wiberg"
        bond_parameter = BondHandler.BondType(
            smirks="[*:1]~[*:2]",
            k=Quantity(1.5, "kilocalorie_per_mole / angstrom**2"),
            length=Quantity(1.5, "angstrom"),
            id="b1000",
        )
        bond_handler.add_parameter(bond_parameter.to_dict())

        forcefield = ForceField()
        forcefield.register_parameter_handler(bond_handler)
        bond_potentials = SMIRNOFFBondCollection.create(
            parameter_handler=forcefield["Bonds"],
            topology=water.to_topology(),
        )

        top_key = BondKey(atom_indices=(0, 1))
        pot_key = bond_potentials.key_map[top_key]
        assert pot_key.associated_handler == "Bonds"
        pot = bond_potentials.potentials[pot_key]

        assert pot.parameters["k"].to(kcal_mol_a2).magnitude == pytest.approx(1.5)

    def test_angle_collection(self):
        angle_handler = AngleHandler(version=0.3)
        angle_parameter = AngleHandler.AngleType(
            smirks="[*:1]~[*:2]~[*:3]",
            k=Quantity(2.5, "kilocalorie_per_mole / radian**2"),
            angle=Quantity(100, "degree"),
            id="b1000",
        )
        angle_handler.add_parameter(angle_parameter.to_dict())

        forcefield = ForceField()
        forcefield.register_parameter_handler(angle_handler)
        angle_potentials = SMIRNOFFAngleCollection.create(
            parameter_handler=forcefield["Angles"],
            topology=Molecule.from_smiles("CCC").to_topology(),
        )

        top_key = AngleKey(atom_indices=(0, 1, 2))
        pot_key = angle_potentials.key_map[top_key]

        assert pot_key.associated_handler == "Angles"

        pot = angle_potentials.potentials[pot_key]

        assert pot.parameters["k"].to(kcal_mol_rad2).magnitude == pytest.approx(2.5)

    def test_store_improper_torsion_matches(self, formaldehyde):
        parameter_handler = ImproperTorsionHandler(version=0.3)
        parameter_handler.add_parameter(
            parameter=ImproperTorsionHandler.ImproperTorsionType(
                smirks="[*:1]~[#6X3:2](~[*:3])~[*:4]",
                periodicity1=2,
                phase1=Quantity(180.0, "degree"),
                k1=Quantity(1.1, "kilocalorie_per_mole"),
            ),
        )

        collection = SMIRNOFFImproperTorsionCollection()
        collection.store_matches(parameter_handler, formaldehyde.to_topology())

        assert len(collection.key_map) == 3

        for indices in [
            (0, 1, 2, 3),
            (0, 2, 3, 1),
            (0, 3, 1, 2),
        ]:
            key = ImproperTorsionKey(atom_indices=indices, mult=0)
            assert key in collection.key_map

    def test_store_nonstandard_improper_idivf(self, acetaldehyde):
        handler = ImproperTorsionHandler(version=0.3)
        handler.add_parameter(
            {
                "smirks": "[*:1]~[#6:2](~[#8:3])~[*:4]",
                "periodicity1": 2,
                "phase1": Quantity(180.0, "degree"),
                "k1": Quantity(1.1, "kilocalorie_per_mole"),
                "idivf1": Quantity(1.234, "dimensionless"),
                "id": "i1",
            },
        )

        collection = SMIRNOFFImproperTorsionCollection.create(
            parameter_handler=handler,
            topology=acetaldehyde.to_topology(),
        )

        handler = ImproperTorsionHandler(version=0.3)
        handler.default_idivf = 5.555
        handler.add_parameter(
            {
                "smirks": "[*:1]~[#6:2](~[#8:3])~[*:4]",
                "periodicity1": 2,
                "phase1": Quantity(180.0, "degree"),
                "k1": Quantity(1.1, "kilocalorie_per_mole"),
                "id": "i1",
            },
        )

        collection = SMIRNOFFImproperTorsionCollection.create(
            parameter_handler=handler,
            topology=acetaldehyde.to_topology(),
        )

        assert Quantity(5.555, "dimensionless") == next(iter(collection.potentials.values())).parameters["idivf"]


class TestBondCollection:
    def test_upconvert_warning(self, parsley, ethanol):
        from packaging.version import Version

        assert parsley["Bonds"].version == Version("0.3")
        with pytest.warns(
            UserWarning,
            match="Automatically up-converting BondHandler from version 0.3 to 0.4.",
        ):
            _create_interchange(parsley, [ethanol])


class TestConstraintCollection:
    @pytest.mark.parametrize(
        ("mol", "n_constraints"),
        [
            ("C", 4),
            ("CC", 6),
        ],
    )
    def test_num_constraints(self, sage, mol, n_constraints):
        bond_handler = sage["Bonds"]
        constraint_handler = sage["Constraints"]

        topology = Molecule.from_smiles(mol).to_topology()

        constraints = SMIRNOFFConstraintCollection.create(
            parameter_handler=[bond_handler, constraint_handler],
            topology=topology,
            bonds=SMIRNOFFBondCollection.create(bond_handler, topology),
        )

        assert len(constraints.key_map) == n_constraints

    def test_constraints_with_distance(self, tip3p, water):
        topology = water.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        constraints = SMIRNOFFConstraintCollection.create(
            parameter_handler=tip3p["Constraints"],
            topology=topology,
        )

        assert len(constraints.key_map) == 3
        assert len(constraints.potentials) == 2


class TestBondOrderInterpolation:
    def test_input_bond_orders_ignored(self, ethanol, xml_ff_bo_bonds):
        """Test that conformers existing in the topology are not considered in the bond order interpolation
        part of the parametrization process"""

        mol = ethanol
        mol.assign_fractional_bond_orders(bond_order_model="am1-wiberg")
        mod_mol = Molecule(mol)
        for bond in mod_mol.bonds:
            bond.fractional_bond_order += 0.1

        top = Topology.from_molecules(mol)
        mod_top = Topology.from_molecules(mod_mol)

        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo_bonds,
        )

        bonds = SMIRNOFFBondCollection.create(
            parameter_handler=forcefield["Bonds"],
            topology=top,
        )
        bonds_mod = SMIRNOFFBondCollection.create(
            parameter_handler=forcefield["Bonds"],
            topology=mod_top,
        )

        for pot_key1, pot_key2 in zip(
            bonds.key_map.values(),
            bonds_mod.key_map.values(),
        ):
            k1 = bonds.potentials[pot_key1].parameters["k"].m_as(kcal_mol_a2)
            k2 = bonds_mod.potentials[pot_key2].parameters["k"].m_as(kcal_mol_a2)
            assert k1 == pytest.approx(k2, rel=1e-5), (k1, k2)

    @skip_if_missing("openmm")
    def test_input_conformers_ignored(self, xml_ff_bo_bonds):
        """Test that conformers existing in the topology are not considered in the bond order interpolation
        part of the parametrization process"""
        from openff.toolkit._tests.test_forcefield import create_ethanol

        mol = create_ethanol()
        mol.assign_fractional_bond_orders(bond_order_model="am1-wiberg")
        mod_mol = Molecule(mol)
        mod_mol.generate_conformers()
        tmp = mod_mol._conformers[0][0][0]
        mod_mol._conformers[0][0][0] = mod_mol._conformers[0][1][0]
        mod_mol._conformers[0][1][0] = tmp

        top = Topology.from_molecules(mol)
        mod_top = Topology.from_molecules(mod_mol)

        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo_bonds,
        )

        bonds = SMIRNOFFBondCollection.create(
            parameter_handler=forcefield["Bonds"],
            topology=top,
        )
        bonds_mod = SMIRNOFFBondCollection.create(
            parameter_handler=forcefield["Bonds"],
            topology=mod_top,
        )

        for key1, key2 in zip(bonds.potentials, bonds_mod.potentials):
            k1 = bonds.potentials[key1].parameters["k"].m_as(kcal_mol_a2)
            k2 = bonds_mod.potentials[key2].parameters["k"].m_as(kcal_mol_a2)
            assert k1 == pytest.approx(k2, rel=1e-5), (k1, k2)

    def test_fractional_bondorder_invalid_interpolation_method(
        self,
        ethanol,
        xml_ff_bo_bonds,
    ):
        """
        Ensure that requesting an invalid interpolation method leads to a
        FractionalBondOrderInterpolationMethodUnsupportedError
        """
        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo_bonds,
        )
        forcefield["Bonds"]._fractional_bondorder_interpolation = "invalid method name"

        # TODO: Make this a more descriptive custom exception
        with pytest.raises(ValidationError):
            Interchange.from_smirnoff(forcefield, [ethanol])


class TestParameterInterpolation:
    @pytest.mark.xfail(reason="Not yet implemented using input bond orders")
    def test_bond_order_interpolation(self, ethanol, xml_ff_bo):
        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo,
        )

        ethanol.generate_conformers(n_conformers=1)

        ethanol.bonds[1].fractional_bond_order = 1.5

        top = ethanol.to_topology()

        out = Interchange.from_smirnoff(forcefield, ethanol.to_topology())

        top_key = BondKey(
            atom_indices=(1, 2),
            bond_order=top.get_bond_between(1, 2).bond.fractional_bond_order,
        )

        found_k = out["Bonds"].potentials[out["Bonds"].key_map[top_key]].parameters["k"]
        assert found_k == 300 * kcal_mol_a2

    @pytest.mark.slow
    @pytest.mark.xfail(reason="Not yet implemented using input bond orders")
    def test_bond_order_interpolation_similar_bonds(self, xml_ff_bo):
        """Test that key mappings do not get confused when two bonds having similar SMIRKS matches
        have different bond orders"""
        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo,
        )

        # TODO: Construct manually to avoid relying on atom ordering
        mol = Molecule.from_smiles("C(CCO)O")
        mol.generate_conformers(n_conformers=1)

        mol.bonds[2].fractional_bond_order = 1.5
        mol.bonds[3].fractional_bond_order = 1.2

        top = mol.to_topology()

        out = Interchange.from_smirnoff(forcefield, top)

        bond1_top_key = BondKey(
            atom_indices=(2, 3),
            bond_order=top.get_bond_between(2, 3).bond.fractional_bond_order,
        )
        bond1_pot_key = out["Bonds"].key_map[bond1_top_key]

        bond2_top_key = BondKey(
            atom_indices=(0, 4),
            bond_order=top.get_bond_between(0, 4).bond.fractional_bond_order,
        )
        bond2_pot_key = out["Bonds"].key_map[bond2_top_key]

        assert numpy.allclose(
            out["Bonds"].potentials[bond1_pot_key].parameters["k"],
            Quantity(300.0, "kilocalories / mol / angstrom ** 2"),
        )

        assert numpy.allclose(
            out["Bonds"].potentials[bond2_pot_key].parameters["k"],
            Quantity(180.0, "kilocalories / mol / angstrom ** 2"),
        )

    @skip_if_missing("openmm")
    @requires_openeye
    @pytest.mark.parametrize(
        (
            "reversed",
            "k_torsion_interpolated",
            "k_bond_interpolated",
            "length_bond_interpolated",
            "central_atoms",
        ),
        [
            (
                False,
                4.165575439001892,
                42207.73245725521,
                0.14005504469705454,
                (1, 2),
            ),
            (
                True,
                4.165322743473034,
                42207.037544550854,
                0.14005579964306572,
                (7, 6),
            ),
        ],
    )
    def test_fractional_bondorder_from_molecule(
        self,
        xml_ff_bo,
        reversed,
        k_torsion_interpolated,
        k_bond_interpolated,
        length_bond_interpolated,
        central_atoms,
        ethanol,
        reversed_ethanol,
    ):
        """Copied from the toolkit with modified reference constants.
        Force constant computed by interpolating (k1, k2) = (101, 123) kcal/A**2/mol
        with bond order 1.00093035 (AmberTools 21.4, Python 3.8, macOS):
            101 + (123 - 101) * (0.00093035) = 101.0204677 kcal/A**2/mol
            = 42266.9637 kJ/nm**2/mol

        Same process with bond length (1.4, 1.3) A gives 0.1399906965 nm
        Same process with torsion k (1.0, 1.8) kcal/mol gives 4.18711406752 kJ/mol

        Using OpenEye (openeye-toolkits 2023.2, Python 3.11, macOS, January 2024):
            bond order 0.9944955302945424
            bond k = 42207.73245725521 kJ/nm**2/mol
            bond length = 0.14005504469705454 nm
            torison k = 4.165575439001892 kj/mol

        ... except OpenEye has a different fractional bond order for reversed ethanol
            bond order 0.9944200356934255
            bond k = 42207.037544550854 kJ/nm**2/mol
            bond length = 0.14005579964306572 nm
            torsion k = 4.165322743473034 kJ/mol

        """
        mol = reversed_ethanol if reversed else ethanol

        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo,
        )
        topology = Topology.from_molecules(mol)

        out = Interchange.from_smirnoff(forcefield, topology)
        out.box = Quantity(4 * numpy.eye(3), "nanometer")
        omm_system = out.to_openmm(combine_nonbonded_forces=True)

        # Verify that the assigned bond parameters were correctly interpolated
        off_bond_force = next(force for force in omm_system.getForces() if isinstance(force, openmm.HarmonicBondForce))

        for idx in range(off_bond_force.getNumBonds()):
            params = off_bond_force.getBondParameters(idx)

            atom1, atom2 = params[0], params[1]
            atom1_mol, atom2_mol = central_atoms

            if ((atom1 == atom1_mol) and (atom2 == atom2_mol)) or ((atom1 == atom2_mol) and (atom2 == atom1_mol)):
                k = params[-1]
                length = params[-2]
                numpy.testing.assert_allclose(
                    k / k.unit,
                    k_bond_interpolated,
                    atol=0,
                    rtol=2e-6,
                )
                numpy.testing.assert_allclose(
                    length / length.unit,
                    length_bond_interpolated,
                    atol=0,
                    rtol=2e-6,
                )

        # Verify that the assigned torsion parameters were correctly interpolated
        off_torsion_force = next(
            force for force in omm_system.getForces() if isinstance(force, openmm.PeriodicTorsionForce)
        )

        for idx in range(off_torsion_force.getNumTorsions()):
            params = off_torsion_force.getTorsionParameters(idx)

            atom2, atom3 = params[1], params[2]
            atom2_mol, atom3_mol = central_atoms

            if ((atom2 == atom2_mol) and (atom3 == atom3_mol)) or ((atom2 == atom3_mol) and (atom3 == atom2_mol)):
                k = params[-1]
                numpy.testing.assert_allclose(
                    k / k.unit,
                    k_torsion_interpolated,
                    atol=0,
                    rtol=2e-6,
                )


def test_get_uses_interpolation():
    handler_no_interpolation = BondHandler(version=0.4)
    handler_all_interpolation = BondHandler(version=0.4)
    handler_some_interpolation = BondHandler(version=0.4)
    handler_partial_interpolation = BondHandler(version=0.4)

    for handler in [
        handler_no_interpolation,
        handler_some_interpolation,
    ]:
        handler.add_parameter(
            {
                "smirks": "[#1:1]-[*:2]",
                "k": Quantity(400, "kilocalories / mol / angstrom ** 2"),
                "length": Quantity(1, "angstrom"),
            },
        )

    for handler in [
        handler_all_interpolation,
        handler_some_interpolation,
    ]:
        handler.add_parameter(
            {
                "smirks": "[#6:1]-[#6:2]",
                "k_bondorder1": Quantity(400, "kilocalories / mol / angstrom ** 2"),
                "length_bondorder1": Quantity(1, "angstrom"),
                "k_bondorder2": Quantity(500, "kilocalories / mol / angstrom ** 2"),
                "length_bondorder2": Quantity(1.2, "angstrom"),
            },
        )

    handler_partial_interpolation.add_parameter(
        {
            "smirks": "[#6:1]-[#7:2]",
            "k": Quantity(400, "kilocalories / mol / angstrom ** 2"),
            "length_bondorder1": Quantity(1, "angstrom"),
            "length_bondorder2": Quantity(0.5, "angstrom"),
        },
    )

    assert not SMIRNOFFBondCollection()._get_uses_interpolation(
        handler_no_interpolation,
    )
    assert SMIRNOFFBondCollection()._get_uses_interpolation(handler_some_interpolation)
    assert SMIRNOFFBondCollection()._get_uses_interpolation(handler_all_interpolation)
    assert SMIRNOFFBondCollection()._get_uses_interpolation(
        handler_partial_interpolation,
    )


def test_check_molecule_uniqueness(
    ethanol,
    reversed_ethanol,
    cyclohexane,
    water,
):
    _check_molecule_uniqueness(None)
    _check_molecule_uniqueness(list())
    _check_molecule_uniqueness([ethanol])
    _check_molecule_uniqueness([cyclohexane])
    _check_molecule_uniqueness([ethanol, cyclohexane, water])

    with pytest.raises(DuplicateMoleculeError, match="Duplicate molecules"):
        _check_molecule_uniqueness(2 * [ethanol])

    with pytest.raises(DuplicateMoleculeError, match="Duplicate molecules"):
        _check_molecule_uniqueness([ethanol, reversed_ethanol])



import random

import pytest
from openff.toolkit import Molecule, Quantity, Topology

from openff.interchange._tests import MoleculeWithConformer
from openff.interchange.exceptions import NegativeMassError


@pytest.mark.parametrize("reversed", [False, True])
def test_hmr_basic(sage, reversed, ethanol, reversed_ethanol):
    pytest.importorskip("openmm.unit")
    import openmm.unit

    hydrogen_mass = random.uniform(1.0, 4.0)

    molecule = reversed_ethanol if reversed else ethanol
    molecule.generate_conformers(n_conformers=1)

    topology = molecule.to_topology()

    interchange = sage.create_interchange(topology)

    system = interchange.to_openmm(hydrogen_mass=hydrogen_mass)

    expected_mass = sum([atom.mass for atom in topology.atoms]).m_as("dalton")

    found_mass = sum(
        [
            system.getParticleMass(particle_index).value_in_unit(openmm.unit.dalton)
            for particle_index in range(system.getNumParticles())
        ],
    )

    assert found_mass == pytest.approx(expected_mass)

    for particle_index, atom in enumerate(topology.atoms):
        if atom.atomic_number == 1:
            assert system.getParticleMass(particle_index)._value == hydrogen_mass


def test_hmr_not_applied_to_water(sage, water):
    # TODO: This should have different behavior for rigid and flexible water,
    #       but sage has tip3p (rigid) so it should always be skipped

    pytest.importorskip("openmm.unit")
    import openmm.unit
    from openmm.app import element

    hydrogen_mass = 1.23

    interchange = sage.create_interchange(water.to_topology())

    system = interchange.to_openmm(hydrogen_mass=hydrogen_mass)

    expected_mass = sum([atom.mass for atom in interchange.topology.atoms]).m_as("dalton")

    found_mass = sum(
        [
            system.getParticleMass(particle_index).value_in_unit(openmm.unit.dalton)
            for particle_index in range(system.getNumParticles())
        ],
    )

    assert found_mass == pytest.approx(expected_mass)

    for particle_index, atom in enumerate(interchange.topology.atoms):
        if atom.atomic_number == 1:
            assert system.getParticleMass(particle_index) == element.hydrogen.mass


def test_mass_is_positive(sage):
    pytest.importorskip("openmm")

    with pytest.raises(
        NegativeMassError,
        match=r"Particle with index 0 would have a negative mass after.*5.0",
    ):
        sage.create_interchange(Molecule.from_smiles("C").to_topology()).to_openmm(hydrogen_mass=5.0)


@pytest.mark.parametrize("hydrogen_mass", [1.1, 3.0])
def test_hmr_not_applied_to_tip4p(
    tip4p,
    water,
    hydrogen_mass,
):
    pytest.importorskip("openmm")

    system = tip4p.create_interchange(
        Topology.from_molecules(2 * [water]),
    ).to_openmm_system(
        hydrogen_mass=hydrogen_mass,
    )

    masses = [system.getParticleMass(particle_index)._value for particle_index in range(system.getNumParticles())]

    assert masses[0] == masses[3] == 15.99943
    assert masses[1] == masses[2] == masses[4] == masses[5] == 1.007947
    assert masses[6] == masses[7] == 0.0


@pytest.mark.parametrize("hydrogen_mass", [1.1, 3.0])
def test_hmr_with_ligand_virtual_sites(
    sage_with_bond_charge,
    hydrogen_mass,
):
    """Test that a single-molecule sigma hole example runs"""
    pytest.importorskip("openmm")
    import openmm
    import openmm.unit

    topology = MoleculeWithConformer.from_mapped_smiles(
        "[H:5][C:2]([H:6])([C:3]([H:7])([H:8])[Cl:4])[Cl:1]",
    ).to_topology()
    topology.box_vectors = Quantity([4, 4, 4], "nanometer")

    # should work fine with 4 fs, but be conservative and just use 3 fs
    simulation = sage_with_bond_charge.create_interchange(
        topology,
    ).to_openmm_simulation(
        integrator=openmm.LangevinMiddleIntegrator(
            300 * openmm.unit.kelvin,
            1 / openmm.unit.picosecond,
            3.0 * openmm.unit.femtosecond,
        ),
        hydrogen_mass=hydrogen_mass,
    )

    system = simulation.system

    assert system.getParticleMass(0)._value == 35.4532
    assert system.getParticleMass(3)._value == 35.4532

    assert system.getParticleMass(1)._value == 12.01078 - 2 * (hydrogen_mass - 1.007947)
    assert system.getParticleMass(2)._value == 12.01078 - 2 * (hydrogen_mass - 1.007947)

    assert system.getParticleMass(4)._value == hydrogen_mass
    assert system.getParticleMass(5)._value == hydrogen_mass
    assert system.getParticleMass(6)._value == hydrogen_mass
    assert system.getParticleMass(7)._value == hydrogen_mass

    assert system.getParticleMass(8)._value == 0.0
    assert system.getParticleMass(9)._value == 0.0

    # make sure it can minimize and run briefly without crashing
    # this is a single ligand in vacuum with boring parameters
    simulation.minimizeEnergy(maxIterations=100)

    simulation.runForClockTime(10 * openmm.unit.second)



from collections.abc import Iterable
from typing import Literal

from openff.toolkit.topology import Atom
from pydantic import Field

from openff.interchange.components.potentials import Collection


class ConstraintCollection(Collection):
    """Collection storing constraint potentials as produced by a SMIRNOFF force field."""

    type: Literal["Constraints"] = "Constraints"
    expression: Literal[""] = ""

    @classmethod
    def potential_parameters(cls) -> Iterable[str]:
        """Return a list of names of parameters included in each potential in this colletion."""
        return ("length",)


class BondCollection(Collection):
    """Collection storing bond potentials."""

    type: Literal["Bonds"] = "Bonds"
    expression: Literal["k/2*(r-length)**2"] = "k/2*(r-length)**2"

    @classmethod
    def potential_parameters(cls) -> Iterable[str]:
        """Return a list of names of parameters included in each potential in this colletion."""
        return "k", "length"

    @classmethod
    def valence_terms(cls, topology) -> list[tuple["Atom", ...]]:
        """Return all bonds in this topology."""
        return [tuple(b.atoms) for b in topology.bonds]


class AngleCollection(Collection):
    """Collection storing Angle potentials."""

    type: Literal["Angles"] = "Angles"
    expression: Literal["k/2*(theta-angle)**2"] = "k/2*(theta-angle)**2"

    @classmethod
    def potential_parameters(cls) -> Iterable[str]:
        """Return a list of names of parameters included in each potential in this colletion."""
        return "k", "angle"

    @classmethod
    def valence_terms(cls, topology):
        """Return all angles in this topology."""
        return [angle for angle in topology.angles]


class ProperTorsionCollection(Collection):
    """Handler storing periodic proper torsion potentials."""

    type: Literal["ProperTorsions"] = "ProperTorsions"
    expression: Literal["k*(1+cos(periodicity*theta-phase))"] = "k*(1+cos(periodicity*theta-phase))"

    @classmethod
    def supported_parameters(cls) -> Iterable[str]:
        """Return a list of supported parameter attribute names."""
        return "k", "periodicity", "phase"


class RyckaertBellemansTorsionCollection(Collection):
    """Handler storing Ryckaert-Bellemans torsion potentials."""

    type: Literal["RBTorsions"] = "RBTorsions"
    expression: str = Field(
        "c0 + "
        "c1 * (cos(phi - 180)) "
        "c2 * (cos(phi - 180)) ** 2 + "
        "c3 * (cos(phi - 180)) ** 3 + "
        "c4 * (cos(phi - 180)) ** 4 + "
        "c5 * (cos(phi - 180)) ** 5",
    )

    @classmethod
    def supported_parameters(cls) -> Iterable[str]:
        """Return a list of supported parameter attribute names."""
        return "c0", "c1", "c2", "c3", "c4", "c5"


class ImproperTorsionCollection(Collection):
    """Handler storing periodic improper torsion potentials."""

    type: Literal["ImproperTorsions"] = "ImproperTorsions"
    expression: Literal["k*(1+cos(periodicity*theta-phase))"] = "k*(1+cos(periodicity*theta-phase))"

    @classmethod
    def supported_parameters(cls) -> Iterable[str]:
        """Return a list of supported parameter attribute names."""
        return "k", "periodicity", "phase"



"""Functions for running energy evluations with all available engines."""

import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING

from openff.utilities.utilities import requires_package

from openff.interchange import Interchange
from openff.interchange.drivers.amber import get_amber_energies
from openff.interchange.drivers.gromacs import get_gromacs_energies
from openff.interchange.drivers.lammps import get_lammps_energies
from openff.interchange.drivers.openmm import get_openmm_energies
from openff.interchange.drivers.report import EnergyReport
from openff.interchange.exceptions import (
    AmberError,
    GMXError,
    LAMMPSError,
    UnsupportedCutoffMethodError,
)

if TYPE_CHECKING:
    import pandas


def get_all_energies(
    interchange: "Interchange",
    combine_nonbonded_forces: bool = False,
    _engines: Iterable[str] = ("OpenMM", "Amber", "GROMACS", "LAMMPS"),
) -> dict[str, EnergyReport]:
    """
    Given an Interchange object, return single-point energies as computed by all available engines.
    """
    # TODO: Return something nan-like if one driver fails, but still return others that succeed
    # TODO: Have each driver return the version of the engine that was used

    all_energies: dict[str, EnergyReport] = dict()

    try:
        # TODO: Worth wiring this argument up to this function? kwargs complexity is not fun
        all_energies["OpenMM"] = get_openmm_energies(
            interchange,
            combine_nonbonded_forces=combine_nonbonded_forces,
        )
    except UnsupportedCutoffMethodError as error:
        warnings.warn(
            f"Skipping OpenMM, driver failed with error:\n\t{error}",
        )

    for engine_name, engine_driver, engine_exception in [
        ("Amber", get_amber_energies, AmberError),
        ("GROMACS", get_gromacs_energies, GMXError),
        ("LAMMPS", get_lammps_energies, LAMMPSError),
    ]:
        if engine_name not in _engines:
            continue
        try:
            all_energies[engine_name] = engine_driver(interchange=interchange)  # type: ignore[operator]
        except engine_exception:
            pass

    return all_energies


@requires_package("pandas")
def get_summary_data(
    interchange: "Interchange",
    combine_nonbonded_forces: bool = False,
    _engines: Iterable[str] = ("OpenMM", "Amber", "GROMACS", "LAMMPS"),
) -> "pandas.DataFrame":
    """Return a pandas DataFrame with summaries of energies from all available engines."""
    from pandas import DataFrame

    energies = get_all_energies(
        interchange,
        combine_nonbonded_forces=combine_nonbonded_forces,
        _engines=_engines,
    )

    for k, v in energies.items():
        for kk in v.energies:
            energies[k].energies[kk] = energies[k].energies[kk].m_as("kilojoule / mole")  # type: ignore[union-attr]

    return DataFrame({k: v.energies for k, v in energies.items()}).T



import packaging.version


def get_lammps_version() -> packaging.version.Version:
    """Convert LAMMPS's YYYYMMDD version string to a PEP 440-compliant version."""
    import lammps

    YYYYMMDD = str(lammps.lammps(cmdargs=["-screen", "none", "-nocite"]).version())

    YYYY = YYYYMMDD[:4]
    MM = YYYYMMDD[4:6]
    DD = YYYYMMDD[6:8]

    return packaging.version.Version(f"{YYYY}.{MM}.{DD}")



"""Unit tests for common virtual site functions."""

from random import random

import numpy
import pytest
from openff.toolkit import ForceField, Quantity

from openff.interchange._tests import MoleculeWithConformer
from openff.interchange.exceptions import MissingVirtualSitesError
from openff.interchange.interop._virtual_sites import get_positions_with_virtual_sites


@pytest.fixture
def tip4p_interchange(water, tip4p):
    return tip4p.create_interchange(water.to_topology())


def test_short_circuit_emtpy_virtual_site_collection(tip3p, water):
    from openff.interchange.smirnoff import SMIRNOFFVirtualSiteCollection

    out = tip3p.create_interchange(water.to_topology())

    with pytest.raises(MissingVirtualSitesError):
        get_positions_with_virtual_sites(out)

    out.collections["VirtualSites"] = SMIRNOFFVirtualSiteCollection()

    with pytest.raises(MissingVirtualSitesError):
        get_positions_with_virtual_sites(out)


def test_nonzero_positions(tip4p_interchange):
    assert not numpy.allclose(
        get_positions_with_virtual_sites(tip4p_interchange)[-1, :].m,
        numpy.zeros((1, 3)),
    )


def test_collate_virtual_site_positions(tip4p, water_dimer):
    out = tip4p.create_interchange(water_dimer)

    out.box = Quantity([10, 10, 10], "nanometer")

    # move the second water far away, since we're later checking that
    # each water's virtual site is close to its oxygen
    out.positions[3:] += Quantity([3, 3, 3], "nanometer")

    positions = get_positions_with_virtual_sites(
        out,
        collate=False,
    ).m_as("nanometer")

    collated_positions = get_positions_with_virtual_sites(
        out,
        collate=True,
    ).m_as("nanometer")

    # first three atoms and last virtual site should not be affected
    assert positions[:3, :] == pytest.approx(collated_positions[:3, :])
    assert positions[-1, :] == pytest.approx(collated_positions[-1, :])

    # first molecule's virtual site is placed at 4th position if collated,
    # second-to-last position if not
    assert positions[-2] == pytest.approx(collated_positions[3])

    def are_close(a, b):
        """Given two positions, return that they're < 1 nanometer apart."""
        return numpy.linalg.norm(a - b) < 1

    # each molecule's oxygen and virtual site should be close-ish
    assert are_close(positions[0], positions[-2])
    assert are_close(positions[3], positions[-1])
    assert are_close(collated_positions[0], collated_positions[3])
    assert are_close(collated_positions[4], collated_positions[7])

    # different molecules' oxygen and virtual site should NOT be close-ish
    assert not are_close(positions[0], positions[-1])
    assert not are_close(positions[3], positions[-2])
    assert not are_close(collated_positions[0], collated_positions[7])
    assert not are_close(collated_positions[4], collated_positions[3])


class TestVirtualSitePositions:
    @pytest.mark.skip(reason="Broken")
    @pytest.mark.parametrize(
        "distance_",
        [
            0.08,
            0.0,
            -0.08,
            -0.16,
            -0.24,
        ],
    )
    def test_bond_charge_positions(
        self,
        sage_with_bond_charge,
        distance_,
    ):
        sage_with_bond_charge["VirtualSites"].get_parameter(
            parameter_attrs={"smirks": "[#6:2]-[#17X1:1]"},
        )[0].distance = Quantity(
            distance_,
            "nanometer",
        )

        out = sage_with_bond_charge.create_interchange(
            MoleculeWithConformer.from_mapped_smiles(
                "[H:3][C:2]([H:4])([H:5])[Cl:1]",
            ).to_topology(),
        )

        positions = get_positions_with_virtual_sites(out)

        distance = numpy.linalg.norm(positions[-1, :].m - positions[0, :].m)

        assert numpy.allclose(
            distance,
            abs(distance_),
        )

    @pytest.mark.parametrize(
        (
            "distance_",
            "theta",
        ),
        [
            (0.08, 120),
            (0.15, 120),
            (0.15, 100),
            (0.06, 180),
            (0.05 + random(), 90 + 90 * random()),
        ],
    )
    def test_planar_monovalent_positions(
        self,
        sage_with_planar_monovalent_carbonyl,
        carbonyl_planar,
        distance_,
        theta,
    ):
        sage_with_planar_monovalent_carbonyl["VirtualSites"].get_parameter(
            parameter_attrs={"smirks": "[#8:1]=[#6X3+0:2]-[#6:3]"},
        )[0].distance = Quantity(
            distance_,
            "nanometer",
        )

        sage_with_planar_monovalent_carbonyl["VirtualSites"].get_parameter(
            parameter_attrs={"smirks": "[#8:1]=[#6X3+0:2]-[#6:3]"},
        )[0].inPlaneAngle = Quantity(
            theta,
            "degree",
        )

        out = sage_with_planar_monovalent_carbonyl.create_interchange(
            carbonyl_planar.to_topology(),
        )

        assert theta == next(iter(out["VirtualSites"].potentials.values())).parameters["inPlaneAngle"].m_as(
            "degree",
        )
        assert distance_ == next(iter(out["VirtualSites"].potentials.values())).parameters["distance"].m_as(
            "nanometer",
        )

        positions = get_positions_with_virtual_sites(out)

        distance = numpy.linalg.norm(
            positions[-1, :].m_as("nanometer") - positions[0, :].m_as("nanometer"),
        )

        try:
            assert distance == pytest.approx(distance_)
        except AssertionError:
            # TODO: Fix me!
            pytest.xfail()

    @pytest.mark.parametrize(
        (
            "distance_",
            "w1",
            "w2",
            "w3",
        ),
        [
            (0.0, 1.0, 0.0, 0.0),
            # TIP4P-FB
            (
                -0.010527445756662016,
                0.8203146574531,
                0.08984267127345003,
                0.08984267127345003,
            ),
            # TIP4P-FB but "backwards"
            (
                0.010527445756662016,
                1.1796853425469,
                -0.08984267127345003,
                -0.08984267127345003,
            ),
            # virtual site place at midpoint
            # d = -1 * (d_OH ** 2 - (0.5 * d_HH) ** 2) ** 0.5
            (-0.0585882276619988, 0.0, 0.5, 0.5),
        ],
    )
    def test_four_site_water_positions(
        self,
        water_tip4p,
        distance_,
        w1,
        w2,
        w3,
    ):
        tip4p = ForceField("tip4p_fb.offxml")

        tip4p["VirtualSites"].get_parameter(
            parameter_attrs={"smirks": "[#1:2]-[#8X2H2+0:1]-[#1:3]"},
        )[0].distance = Quantity(
            distance_,
            "nanometer",
        )

        out = tip4p.create_interchange(water_tip4p.to_topology())

        positions = get_positions_with_virtual_sites(out)

        p1 = out.positions[0]
        p2 = out.positions[1]
        p3 = out.positions[2]

        assert numpy.allclose(
            positions[-1].m_as("angstrom"),
            (w1 * p1 + w2 * p2 + w3 * p3).m_as("angstrom"),
        )

    @pytest.mark.parametrize(
        "distance_",
        [
            0.8,
            0.0,
            -0.8,
            -1.6,
            -2.4,
        ],
    )
    def test_trivalent_nitrogen_positions(
        self,
        sage_with_trivalent_nitrogen,
        ammonia_tetrahedral,
        distance_,
    ):
        sage_with_trivalent_nitrogen["VirtualSites"].get_parameter(
            parameter_attrs={"smirks": "[#1:2][#7:1]([#1:3])[#1:4]"},
        )[0].distance = Quantity(
            distance_,
            "angstrom",
        )

        out = sage_with_trivalent_nitrogen.create_interchange(
            ammonia_tetrahedral.to_topology(),
        )

        positions = get_positions_with_virtual_sites(out).to("angstrom")

        distance = numpy.linalg.norm(positions[-1, :].m - positions[0, :].m)

        assert distance == pytest.approx(abs(distance_))

        # The nitrogen is placed at [0, 0, 0.8855572013] and the hydrogens are on
        # the xy plane, so the virtual site is at [0, 0, 0.88 ... + distance_]
        assert positions[-1].m_as("angstrom")[0] == pytest.approx(0.0)
        # assert positions[-1].m_as("angstrom")[1] == pytest.approx(0.0)



import random

import pytest

from openff.interchange.exceptions import UnsupportedExportError


@pytest.mark.parametrize("reversed", [False, True])
def test_hmr_basic(sage, reversed, ethanol, reversed_ethanol):
    parmed = pytest.importorskip("parmed")
    hydrogen_mass = random.uniform(1.0, 4.0)

    molecule = reversed_ethanol if reversed else ethanol
    molecule.generate_conformers(n_conformers=1)

    topology = molecule.to_topology()

    interchange = sage.create_interchange(topology)
    interchange.box = [10, 10, 10]

    interchange.to_gromacs(prefix="asdf", hydrogen_mass=hydrogen_mass)

    structure = parmed.load_file("asdf.top")

    expected_mass = sum([atom.mass for atom in topology.atoms]).m_as("dalton")

    found_mass = sum([atom.mass for atom in structure.atoms])

    assert found_mass == pytest.approx(expected_mass)

    for particle_index, atom in enumerate(topology.atoms):
        if atom.atomic_number == 1:
            assert structure.atoms[particle_index].mass == pytest.approx(hydrogen_mass)


def test_hmr_not_applied_to_water(sage, water):
    parmed = pytest.importorskip("parmed")
    # TODO: This should have different behavior for rigid and flexible water,
    #       but sage has tip3p (rigid) so it should always be skipped
    hydrogen_mass = 1.23

    interchange = sage.create_interchange(water.to_topology())
    interchange.box = [10, 10, 10]

    interchange.to_gromacs(prefix="fff", hydrogen_mass=hydrogen_mass)

    structure = parmed.load_file("fff.top")

    expected_mass = sum([atom.mass for atom in water.atoms]).m_as("dalton")

    found_mass = sum([atom.mass for atom in structure.atoms])

    assert found_mass == pytest.approx(expected_mass)

    for particle_index, atom in enumerate(interchange.topology.atoms):
        if atom.atomic_number == 1:
            assert structure.atoms[particle_index].mass == pytest.approx(1.007947)


def test_virtual_sites_unsupported(tip4p, water):
    with pytest.raises(UnsupportedExportError, match="with virtual sites"):
        tip4p.create_interchange(water.to_topology()).to_gromacs(
            prefix="no_vs",
            hydrogen_mass=2.0,
        )



from typing import Literal

from openff.toolkit import Quantity, Topology
from pydantic import Field, PrivateAttr

from openff.interchange._annotations import _DistanceQuantity
from openff.interchange.common._nonbonded import ElectrostaticsCollection, vdWCollection
from openff.interchange.components.potentials import Potential
from openff.interchange.foyer._base import _copy_params
from openff.interchange.foyer._guard import has_foyer
from openff.interchange.models import PotentialKey, TopologyKey

if has_foyer:
    try:
        from foyer import Forcefield
        from foyer.atomtyper import find_atomtypes
        from foyer.topology_graph import TopologyGraph
    except ModuleNotFoundError:
        pass


class FoyerVDWHandler(vdWCollection):
    """Handler storing vdW potentials as produced by a Foyer force field."""

    force_field_key: str = "atoms"
    mixing_rule: Literal["lorentz-berthelot", "geometric"] = Field("geometric")
    method: Literal["cutoff", "pme", "no-cutoff"] = Field("cutoff")

    def store_matches(
        self,
        force_field: "Forcefield",
        topology: "Topology",
    ) -> None:
        """Populate self.key_map with key-val pairs of [TopologyKey, PotentialKey]."""

        top_graph = TopologyGraph.from_openff_topology(topology)

        type_map = find_atomtypes(top_graph, forcefield=force_field)
        for key, val in type_map.items():
            top_key = TopologyKey(atom_indices=(key,))
            self.key_map[top_key] = PotentialKey(id=val["atomtype"])

    def store_potentials(self, force_field: "Forcefield") -> None:
        """Extract specific force field potentials a Forcefield object."""
        for top_key in self.key_map:
            atom_params = force_field.get_parameters(
                self.force_field_key,
                key=self.key_map[top_key].id,
            )

            atom_params = _copy_params(
                atom_params,
                "charge",
                param_units={"epsilon": "kJ / mol", "sigma": "nanometer"},
            )

            self.potentials[self.key_map[top_key]] = Potential(parameters=atom_params)


class FoyerElectrostaticsHandler(ElectrostaticsCollection):
    """Handler storing electrostatics potentials as produced by a Foyer force field."""

    force_field_key: str = "atoms"
    cutoff: _DistanceQuantity = Quantity(9.0, "angstrom")

    _charges: dict[TopologyKey, Quantity] = PrivateAttr(dict())

    def store_charges(
        self,
        atom_slots: dict[TopologyKey, PotentialKey],
        force_field: "Forcefield",
    ):
        """Look up fixed charges (a.k.a. library charges) from the force field and store them in self._charges."""
        for top_key, pot_key in atom_slots.items():
            foyer_params = force_field.get_parameters(self.force_field_key, pot_key.id)

            charge = Quantity(foyer_params["charge"], "elementary_charge")

            self._charges[top_key] = charge

            self.key_map[top_key] = pot_key
            self.potentials[pot_key] = Potential(
                parameters={"charge": charge},
            )



import pytest
from openff.toolkit import Topology


@pytest.mark.parametrize("name_sol", [True, False])
@pytest.mark.parametrize("collate", [True, False])
def test_collate_or_not_virtual_site_ordering(
    tip4p,
    water,
    name_sol,
    collate,
):
    pytest.importorskip("openmm")

    if name_sol:
        for atom in water.atoms:
            atom.metadata["residue_name"] = "SOL"

    openmm_topology = tip4p.create_interchange(
        Topology.from_molecules([water, water]),
    ).to_openmm_topology(
        collate=collate,
    )

    if collate:
        # particles should be O H H VS O H H VS
        #                     0 1 2 3  4 5 6 7
        virtual_site_indices = (3, 7)

    else:
        # particles should be O H H O H H VS VS
        #                     0 1 2 3 4 5 6  7
        virtual_site_indices = (6, 7)

    for particle_index, particle in enumerate(openmm_topology.atoms()):
        assert (particle.element is None) == (particle_index in virtual_site_indices)

    assert {residue.name for residue in openmm_topology.residues()} == {"SOL"} if name_sol else {"UNK"}


def test_each_molecule_with_virtual_sites_has_its_own_funnky_virtual_site_residue(
    tip5p,
    water,
):
    topology = Topology.from_molecules([water, water])

    for index, molecule in enumerate(topology.molecules):
        for atom in molecule.atoms:
            atom.metadata["residue_name"] = "SOL"
            atom.metadata["residue_number"] = str(index)

    openmm_topology = tip5p.create_interchange(topology).to_openmm_topology()

    assert openmm_topology.getNumResidues() == 4

    assert [3, 3, 2, 2] == [len([*residue.atoms()]) for residue in openmm_topology.residues()]



import functools
from collections.abc import Callable
from typing import Annotated, Any

import numpy
from annotated_types import Gt
from openff.toolkit import Quantity
from pydantic import (
    AfterValidator,
    BeforeValidator,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapSerializer,
    WrapValidator,
)

PositiveFloat = Annotated[float, Gt(0)]


def _has_compatible_dimensionality(
    quantity: Quantity,
    unit: str,
    convert: bool,
) -> Quantity:
    """Check if a Quantity has the same dimensionality as a given unit and optionally convert."""
    if quantity.is_compatible_with(unit):
        if convert:
            return quantity.to(unit)
        else:
            return quantity
    else:
        raise ValueError(
            f"Dimensionality of {quantity=} is not compatible with {unit=}",
        )


def _dimensionality_validator_factory(unit: str) -> Callable:
    """Return a function, meant to be passed to a validator, that checks for a specific unit."""
    return functools.partial(_has_compatible_dimensionality, unit=unit, convert=False)


def _unit_validator_factory(unit: str) -> Callable:
    """Return a function, meant to be passed to a validator, that checks for a specific unit."""
    return functools.partial(_has_compatible_dimensionality, unit=unit, convert=True)


(
    _is_distance,
    _is_velocity,
    _is_mass,
    _is_temperature,
) = (
    _dimensionality_validator_factory(unit=_unit)
    for _unit in [
        "nanometer",
        "nanometer / picosecond",
        "unified_atomic_mass_unit",
        "kelvin",
    ]
)

(
    _is_dimensionless,
    _is_kj_mol,
    _is_nanometer,
    _is_degree,
    _is_elementary_charge,
) = (
    _unit_validator_factory(unit=_unit)
    for _unit in [
        "dimensionless",
        "kilojoule / mole",
        "nanometer",
        "degree",
        "elementary_charge",
    ]
)


def quantity_validator(
    value: str | Quantity | dict,
    handler: ValidatorFunctionWrapHandler,
    info: ValidationInfo,
) -> Quantity:
    """Take Quantity-like objects and convert them to Quantity objects."""
    if info.mode == "json":
        assert isinstance(value, dict), "Quantity must be in dict form here."

        # this is coupled to how a Quantity looks in JSON
        return Quantity(value["val"], value["unit"])

        # some more work may be needed to work with arrays, lists, tuples, etc.

    assert info.mode == "python"

    if isinstance(value, Quantity):
        return value
    elif isinstance(value, str):
        return Quantity(value)
    elif isinstance(value, dict):
        return Quantity(value["val"], value["unit"])
    if "openmm" in str(type(value)):
        from openff.units.openmm import from_openmm

        return from_openmm(value)
    else:
        raise ValueError(f"Invalid type {type(value)} for Quantity")


def quantity_json_serializer(
    quantity: Quantity,
    nxt,
) -> dict:
    """Serialize a Quantity to a JSON-compatible dictionary."""
    magnitude = quantity.m

    if isinstance(magnitude, numpy.ndarray):
        # This could be something fancier, list a bytestring
        magnitude = magnitude.tolist()

    return {
        "val": magnitude,
        "unit": str(quantity.units),
    }


# Pydantic v2 likes to marry validators and serializers to types with Annotated
# https://docs.pydantic.dev/latest/concepts/validators/#annotated-validators
_Quantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    WrapSerializer(quantity_json_serializer),
]

_DimensionlessQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_dimensionless),
    WrapSerializer(quantity_json_serializer),
]

_DistanceQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_distance),
    WrapSerializer(quantity_json_serializer),
]

_LengthQuantity = _DistanceQuantity

_VelocityQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_velocity),
    WrapSerializer(quantity_json_serializer),
]

_MassQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_mass),
    WrapSerializer(quantity_json_serializer),
]

_TemperatureQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_temperature),
    WrapSerializer(quantity_json_serializer),
]

_DegreeQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_degree),
    WrapSerializer(quantity_json_serializer),
]

_ElementaryChargeQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_elementary_charge),
    WrapSerializer(quantity_json_serializer),
]

_kJMolQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_kj_mol),
    WrapSerializer(quantity_json_serializer),
]


def _is_positions_shape(quantity: Quantity) -> Quantity:
    if quantity.m.shape[1] == 3:
        return quantity
    else:
        raise ValueError(
            f"Quantity {quantity} of wrong shape ({quantity.shape}) to be positions.",
        )


def _duck_to_nanometer(value: Any):
    """Cast list or ndarray without units to Quantity[ndarray] of nanometer."""
    if isinstance(value, list | numpy.ndarray):
        return Quantity(value, "nanometer")
    elif isinstance(value, Quantity):
        return value.to("nanometer")
    else:
        return value


_PositionsQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_nanometer),
    AfterValidator(_is_positions_shape),
    BeforeValidator(_duck_to_nanometer),
    WrapSerializer(quantity_json_serializer),
]


def _is_box_shape(quantity) -> Quantity:
    if quantity.m.shape == (3, 3):
        return quantity
    elif quantity.m.shape == (3,):
        return numpy.eye(3) * quantity
    else:
        raise ValueError(f"Quantity {quantity} is not a box.")


def _unwrap_list_of_openmm_quantities(value: Any):
    """Unwrap a list of OpenMM quantities to a single Quantity."""
    if isinstance(value, list):
        if any(["openmm" in str(type(element)) for element in value]):
            from openff.units.openmm import from_openmm

            if len({element.unit for element in value}) != 1:
                raise ValueError("All units must be the same.")

            return from_openmm(value)

        else:
            return value

    else:
        return value


_BoxQuantity = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_is_distance),
    AfterValidator(_is_box_shape),
    BeforeValidator(_duck_to_nanometer),
    BeforeValidator(_unwrap_list_of_openmm_quantities),
    WrapSerializer(quantity_json_serializer),
]



import copy

import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity
from openff.utilities.testing import skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer
from openff.interchange.drivers import get_openmm_energies
from openff.interchange.exceptions import (
    CutoffMismatchError,
    SwitchingFunctionMismatchError,
    UnsupportedCombinationError,
)
from openff.interchange.warnings import InterchangeCombinationWarning


class TestCombine:
    @skip_if_missing("openmm")
    def test_basic_combination(self, sage_unconstrained):
        """Test basic use of Interchange.__add__() based on the README example"""
        top = MoleculeWithConformer.from_smiles("C").to_topology()

        interchange = Interchange.from_smirnoff(sage_unconstrained, top)

        interchange.box = [4, 4, 4] * numpy.eye(3)

        # Copy and translate atoms by [1, 1, 1]
        other = Interchange(topology=copy.deepcopy(interchange.topology))
        other = copy.deepcopy(interchange)
        other.positions += Quantity("1.0 nanometer")

        combined = interchange.combine(other)

        # Just see if it can be converted into OpenMM and run
        get_openmm_energies(combined)

    @pytest.mark.filterwarnings("ignore:Setting positions to None")
    @pytest.mark.slow
    def test_parameters_do_not_clash(self, monkeypatch, sage_unconstrained):
        thf = MoleculeWithConformer.from_smiles("C1CCOC1")
        ace = MoleculeWithConformer.from_smiles("CC(=O)O")

        def make_interchange(molecule: MoleculeWithConformer) -> Interchange:
            interchange = Interchange.from_smirnoff(
                force_field=sage_unconstrained,
                topology=[molecule],
            )
            interchange.positions = molecule.conformers[0]

            return interchange

        thf_interchange = make_interchange(thf)
        ace_interchange = make_interchange(ace)
        complex_interchange = thf_interchange.combine(ace_interchange)

        thf_vdw = thf_interchange["vdW"].get_system_parameters()
        ace_vdw = ace_interchange["vdW"].get_system_parameters()
        add_vdw = complex_interchange["vdW"].get_system_parameters()

        numpy.testing.assert_equal(numpy.vstack([thf_vdw, ace_vdw]), add_vdw)

        # TODO: Ensure the de-duplication is maintained after exports

    def test_positions_setting(self, sage):
        """Test that positions exist on the result if and only if
        both input objects have positions."""

        ethane = Molecule.from_smiles("CC")
        methane = Molecule.from_smiles("C")

        ethane_interchange = Interchange.from_smirnoff(
            sage,
            [ethane],
        )
        methane_interchange = Interchange.from_smirnoff(sage, [methane])

        ethane.generate_conformers(n_conformers=1)
        methane.generate_conformers(n_conformers=1)

        assert (methane_interchange.combine(ethane_interchange)).positions is None
        methane_interchange.positions = methane.conformers[0]
        assert (methane_interchange.combine(ethane_interchange)).positions is None
        ethane_interchange.positions = ethane.conformers[0]
        assert (methane_interchange.combine(ethane_interchange)).positions is not None

    @pytest.mark.parametrize("handler", ["Electrostatics", "vdW"])
    def test_error_mismatched_cutoffs(
        self,
        sage,
        basic_top,
        handler,
    ):
        sage_modified = ForceField("openff-2.1.0.offxml")

        sage_modified[handler].cutoff *= 1.5

        with pytest.raises(
            CutoffMismatchError,
            match=f"{handler} cutoffs do not match",
        ):
            sage.create_interchange(basic_top).combine(
                sage_modified.create_interchange(basic_top),
            )

    @pytest.mark.parametrize("key", ["Electrostatics", "vdW"])
    def test_cutoffs_only_slightly_differ(
        self,
        sage,
        basic_top,
        key,
    ):
        """Test that, by default, combining proceeds if 1-16 < cutoff_diff < 1e-6."""
        int1 = sage.create_interchange(basic_top)
        int2 = int1.__deepcopy__()

        # should combine with a tiny difference that's still fails x1 == x2, but ...
        int2[key].cutoff = int1[key].cutoff + Quantity("1e-10 nanometer")

        int1.combine(int2)

        # ... fails when the difference is more than 1e-6
        int2[key].cutoff = int1[key].cutoff + Quantity("1e-4 nanometer")

        with pytest.raises(
            CutoffMismatchError,
        ):
            int1.combine(int2)

    def test_error_mismatched_switching_function(
        self,
        sage,
        basic_top,
    ):
        sage_modified = ForceField("openff-2.1.0.offxml")

        sage_modified["vdW"].switch_width *= 0.0

        with pytest.raises(
            SwitchingFunctionMismatchError,
        ):
            sage.create_interchange(basic_top).combine(
                sage_modified.create_interchange(basic_top),
            )

    @pytest.mark.parametrize(
        argnames=["vdw", "electrostatics"],
        argvalues=[
            (True, False),
            (False, True),
            (True, True),
            (False, False),
        ],
    )
    def test_dont_combine_mixed_14(self, sage, vdw, electrostatics):
        """
        Until it's implemented, error out when any non-bonded collections have non-equal 1-4 scaling factors.

        See https://github.com/openforcefield/openff-interchange/issues/380

        """
        interchange1 = sage.create_interchange(MoleculeWithConformer.from_smiles("C").to_topology())
        interchange2 = sage.create_interchange(MoleculeWithConformer.from_smiles("CCO").to_topology())

        if vdw:
            interchange2["vdW"].scale_14 = 0.444

        if electrostatics:
            interchange2["Electrostatics"].scale_14 = 0.444

        if vdw or electrostatics:
            with pytest.raises(
                UnsupportedCombinationError,
                match="1-4.*vdW" if vdw else "1-4.*Electro",
            ):
                interchange2.combine(interchange1)
        else:
            # if neither are modified, that error shouldn't be raised
            interchange2.combine(interchange1)

    def test_mix_different_5_6_rounding(self, parsley, sage, ethanol):
        """Test that 0.833333 is 'upconverted' to 0.8333333333 in combination."""
        with pytest.warns(
            InterchangeCombinationWarning,
            match="more digits in rounding",
        ):
            parsley.create_interchange(
                ethanol.to_topology(),
            ).combine(
                sage.create_interchange(ethanol.to_topology()),
            )

    @pytest.mark.parametrize("flip_order", [False, True])
    def test_no_unnecessary_duplicate_tags(self, water_dimer, tip3p, sage, ethanol, flip_order):
        """Check that the '_DUPLICATE' hack is only used when necessary. See Issue #1324."""
        interchanges = [
            tip3p.create_interchange(
                water_dimer,
            ),
            sage.create_interchange(
                ethanol.to_topology(),
            ),
        ]

        if flip_order:
            interchanges.reverse()

        combined = interchanges[0].combine(interchanges[1])

        for collection_name in ["vdW", "Bonds", "Angles", "ProperTorsions", "Constraints"]:
            for potential_key in combined[collection_name].key_map.values():
                assert "_DUPLICATE" not in potential_key.id, (
                    f"Failed sanity check with {potential_key.id} in {collection_name}"
                )

    @pytest.mark.parametrize("flip_order", [False, True])
    @pytest.mark.parametrize("handler_name", ["Constraints", "Bonds", "Angles", "ProperTorsions"])
    def test_constraint_key_collision(self, parsley, sage, ethanol, flip_order, handler_name):
        """Test that key collisions in constraints and valence terms are handled."""
        interchanges = [
            parsley.create_interchange(
                ethanol.to_topology(),
            ),
            sage.create_interchange(
                ethanol.to_topology(),
            ),
        ]

        # want to make sure this behavior is not order-dependent
        if flip_order:
            interchanges.reverse()

        arrays_before = [interchange[handler_name].get_system_parameters() for interchange in interchanges]

        numpy.testing.assert_raises(AssertionError, numpy.testing.assert_allclose, *arrays_before)

        array_after_combine = interchanges[0].combine(interchanges[1])[handler_name].get_system_parameters()

        # check that the contents of the combined Interchange contains each input, without mushing
        numpy.testing.assert_allclose(
            numpy.vstack(arrays_before),
            array_after_combine,
        )

    @pytest.mark.parametrize("flip_order", [False, True])
    def test_charge_cache_behavior(self, sage, ethanol, caffeine, flip_order):
        int1 = sage.create_interchange(ethanol.to_topology())
        int2 = sage.create_interchange(caffeine.to_topology())

        for interchange_ in (int1, int2):
            assert len(interchange_["Electrostatics"]._get_charges()) == interchange_.topology.n_atoms

        if flip_order:
            int3 = int2.combine(int1)
        else:
            int3 = int1.combine(int2)

        assert int3["Electrostatics"]._charges_cached is False

        # Check that the input caches are intact
        for interchange_ in (int1, int2):
            assert len(interchange_["Electrostatics"]._get_charges()) == interchange_.topology.n_atoms

    def test_same_keys_same_physics_combines_without_error(self, sage, methane, ethanol):
        """Test that PotentialKey collisions are only corrected when the associated parameters differ."""
        ic1 = sage.create_interchange(methane.to_topology())
        ic2 = sage.create_interchange(ethanol.to_topology())

        # should not error no matter the order of combination
        ic3 = ic1.combine(ic2)
        ic4 = ic2.combine(ic1)

        # should not have shim _DUPLICATE keys
        for interchange in [ic3, ic4]:
            for name, collection in interchange.collections.items():
                if name == "Constraints":
                    continue
                for key in collection.potentials.keys():
                    assert "_DUPLICATE" not in key.id, f"Unexpected _DUPLICATE in {key.id}"

        numpy.testing.assert_allclose(
            numpy.vstack(
                [
                    ic1["vdW"].get_system_parameters(),
                    ic2["vdW"].get_system_parameters(),
                ],
            ),
            ic3["vdW"].get_system_parameters(),
        )

        numpy.testing.assert_allclose(
            numpy.vstack(
                [
                    ic2["vdW"].get_system_parameters(),
                    ic1["vdW"].get_system_parameters(),
                ],
            ),
            ic4["vdW"].get_system_parameters(),
        )

    def test_shim_duplicate_key_already_exists(self, parsley, sage, methane, ethanol):
        ic1 = parsley.create_interchange(methane.to_topology())
        ic2 = sage.create_interchange(ethanol.to_topology())

        # should error no matter the order of combination
        with pytest.raises(
            UnsupportedCombinationError,
            match="already have _DUPLICATE appended",
        ):
            ic1.combine(ic2.combine(ic1))

        with pytest.raises(
            UnsupportedCombinationError,
            match="already have _DUPLICATE appended",
        ):
            ic2.combine(ic2.combine(ic1))

    def test_combine_twice(self, sage, methane, ethanol, water_dimer):
        """Test that combining more than two Interchanges works as expected."""

        # another good test would be to do this with three separate FFs, and then ...
        ic1 = sage.create_interchange(methane.to_topology())
        ic2 = sage.create_interchange(ethanol.to_topology())
        ic3 = sage.create_interchange(water_dimer)

        ic4 = ic1.combine(ic2).combine(ic3)
        # ... make ic5 out of ForceField("ff1.offxml", "ff2.offxml", "ff3.offxml")
        # and a combined topology, and compare the parameters (and more?) there

        for collection_name in ["vdW", "Bonds", "Angles", "Constraints"]:
            numpy.testing.assert_allclose(
                numpy.vstack(
                    [
                        ic1[collection_name].get_system_parameters(),
                        ic2[collection_name].get_system_parameters(),
                        ic3[collection_name].get_system_parameters(),
                    ],
                ),
                ic4[collection_name].get_system_parameters(),
            )

        # vstack is basically concatenation in 2D, but these are 1D arrays
        numpy.testing.assert_allclose(
            numpy.concatenate(
                [
                    ic1["Electrostatics"].get_charge_array(),
                    ic2["Electrostatics"].get_charge_array(),
                    ic3["Electrostatics"].get_charge_array(),
                ],
            ),
            ic4["Electrostatics"].get_charge_array(),
        )

    def test_combine_from_openmm_add_nagl_to_openmm(self, sage, ethanol):
        topology = ethanol.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        system1 = sage.create_interchange(topology)

        system2 = Interchange.from_openmm(
            system=system1.to_openmm(),
            topology=system1.topology.to_openmm(),
        )

        system2.combine(system1).to_openmm()



"""Functions for running energy evluations with OpenMM."""

import warnings
from typing import TYPE_CHECKING, Optional

import numpy
from openff.toolkit import Quantity
from openff.units import ensure_quantity
from openff.utilities.utilities import has_package, requires_package

from openff.interchange import Interchange
from openff.interchange.drivers.report import EnergyReport
from openff.interchange.exceptions import CannotInferNonbondedEnergyError
from openff.interchange.interop.openmm._positions import to_openmm_positions

if has_package("openmm") or TYPE_CHECKING:
    import openmm
    import openmm.unit


@requires_package("openmm")
def get_openmm_energies(
    interchange: Interchange,
    round_positions: int | None = None,
    combine_nonbonded_forces: bool = True,
    detailed: bool = False,
    platform: str = "Reference",
) -> EnergyReport:
    """
    Given an OpenFF Interchange object, return single-point energies as computed by OpenMM.

    Parameters
    ----------
    interchange
        An OpenFF Interchange object to compute the single-point energy of
    round_positions
        The number of decimal places, in nanometers, to round positions. This can be useful when
        comparing to i.e. GROMACS energies, in which positions may be rounded.
    combine_nonbonded_forces
        Whether or not to combine all non-bonded interactions (vdW, short- and long-range
        ectrostaelectrostatics, and 1-4 interactions) into a single openmm.NonbondedForce.
    platform
        The name of the platform (`openmm.Platform`) used by OpenMM in this calculation.
    detailed
        Attempt to report energies with more granularity. Requires `combine_nonbonded_forces=False`.
        Not guaranteed to be compatible with non-default values of other arguments. Useful for
        debugging.

    Returns
    -------
    report
        An `EnergyReport` object containing the single-point energies.

    """
    if detailed and combine_nonbonded_forces:
        raise ValueError(
            "A detailed energy report (`detailed=True`) requires non-bonded interactions be "
            "split out into different forces (`combine_nonbonded_forces=False`).",
        )

    if "VirtualSites" in interchange.collections:
        if len(interchange["VirtualSites"].key_map) > 0:
            if not combine_nonbonded_forces:
                warnings.warn(
                    "Collecting energies from split forces with virtual sites is experimental",
                    UserWarning,
                )

            has_virtual_sites = True
        else:
            has_virtual_sites = False
    else:
        has_virtual_sites = False

    system: openmm.System = interchange.to_openmm(
        combine_nonbonded_forces=combine_nonbonded_forces,
    )

    box_vectors: openmm.unit.Quantity = None if interchange.box is None else interchange.box.to_openmm()

    positions: openmm.unit.Quantity = to_openmm_positions(
        interchange,
        include_virtual_sites=has_virtual_sites,
    )

    return _process(
        _get_openmm_energies(
            system=system,
            box_vectors=box_vectors,
            positions=positions,
            round_positions=round_positions,
            platform=platform,
        ),
        combine_nonbonded_forces=combine_nonbonded_forces,
        detailed=detailed,
        system=system,
    )


def _get_openmm_energies(
    system: "openmm.System",
    box_vectors: Optional["openmm.unit.Quantity"],
    positions: "openmm.unit.Quantity",
    round_positions: int | None,
    platform: str,
) -> dict[int, "openmm.unit.Quantity"]:
    """Given prepared `openmm` objects, run a single-point energy calculation."""
    for index, force in enumerate(system.getForces()):
        force.setForceGroup(index)

    integrator = openmm.LangevinMiddleIntegrator(
        300 * openmm.unit.kelvin,
        1.0 / openmm.unit.picosecond,
        1.0 * openmm.unit.femtoseconds,
    )

    context = openmm.Context(
        system,
        integrator,
        openmm.Platform.getPlatformByName(platform),
    )

    if box_vectors is not None:
        context.setPeriodicBoxVectors(*box_vectors)

    context.setPositions(
        (numpy.round(positions, round_positions) if round_positions is not None else positions),
    )

    raw_energies: dict[int, openmm.unit.Quantity] = dict()

    for index in range(system.getNumForces()):
        state = context.getState(getEnergy=True, groups={index})
        raw_energies[index] = state.getPotentialEnergy()
        del state

    del context
    del integrator

    return raw_energies


def _process(
    raw_energies: dict[int, "openmm.unit.Quantity"],
    system: "openmm.System",
    combine_nonbonded_forces: bool,
    detailed: bool,
) -> EnergyReport:
    staged: dict[str, Quantity] = dict()

    valence_map = {
        openmm.HarmonicBondForce: "Bond",
        openmm.HarmonicAngleForce: "Angle",
        openmm.PeriodicTorsionForce: "Torsion",
        openmm.RBTorsionForce: "RBTorsion",
    }

    # This assumes that only custom forces will have duplicate instances
    for index, raw_energy in raw_energies.items():
        force = system.getForce(index)

        if type(force) in valence_map:
            staged[valence_map[type(force)]] = raw_energy

        elif type(force) in [
            openmm.NonbondedForce,
            openmm.CustomNonbondedForce,
            openmm.CustomBondForce,
        ]:
            if combine_nonbonded_forces:
                assert isinstance(force, openmm.NonbondedForce)

                staged["Nonbonded"] = raw_energy

                continue

            else:
                if isinstance(force, openmm.NonbondedForce):
                    staged["Electrostatics"] = raw_energy

                elif isinstance(force, openmm.CustomNonbondedForce):
                    staged["vdW"] = raw_energy

                elif isinstance(force, openmm.CustomBondForce):
                    if "qq" in force.getEnergyFunction():
                        staged["Electrostatics 1-4"] = raw_energy
                    else:
                        staged["vdW 1-4"] = raw_energy

                else:
                    raise CannotInferNonbondedEnergyError()

    if detailed:
        processed = staged

    else:
        processed = {key: staged[key] for key in ["Bond", "Angle", "Torsion", "RBTorsion"] if key in staged}

        nonbonded_energies = [
            staged[key]
            for key in [
                "Nonbonded",
                "Electrostatics",
                "vdW",
                "Electrostatics 1-4",
                "vdW 1-4",
            ]
            if key in staged
        ]

        # Array inference acts up if given a 1-list of Quantity
        if combine_nonbonded_forces:
            assert len(nonbonded_energies) == 1

            processed["Nonbonded"] = nonbonded_energies[0]

        else:
            zero = 0.0 * openmm.unit.kilojoule_per_mole

            processed["Electrostatics"] = ensure_quantity(
                numpy.sum(
                    [staged.get(key, zero) for key in ["Electrostatics", "Electrostatics 1-4"]],
                ),
                "openff",
            )

            processed["vdW"] = ensure_quantity(
                numpy.sum([staged.get(key, zero) for key in ["vdW", "vdW 1-4"]]),
                "openff",
            )

    return EnergyReport(energies=processed)



from collections.abc import Iterable

class NGLWidget:
    def __init__(
        self,
        structure=None,
        representations=None,
        parameters=None,
        **kwargs,
    ): ...
    def add_representation(
        self,
        repr_type: str,
        selection: str | Iterable,
        **kwargs,
    ): ...

def show_file(file_path: str) -> NGLWidget: ...



"""Custom classes exposed as plugins."""

import math
from collections.abc import Iterable
from typing import Literal

from openff.interchange._annotations import _DimensionlessQuantity, _DistanceQuantity
from openff.interchange.components.potentials import Potential
from openff.interchange.exceptions import InvalidParameterHandlerError
from openff.interchange.smirnoff._nonbonded import _SMIRNOFFNonbondedCollection
from openff.toolkit import Quantity, Topology, unit
from openff.toolkit.typing.engines.smirnoff.parameters import (
    ParameterAttribute,
    ParameterHandler,
    ParameterType,
    VirtualSiteHandler,
    _allow_only,
)

_HandlerIterable = Iterable[type[ParameterHandler]]
_CollectionAlias = type[_SMIRNOFFNonbondedCollection]


class BuckinghamHandler(ParameterHandler):
    """A custom SMIRNOFF handler for Buckingham interactions."""

    class BuckinghamType(ParameterType):
        """A custom SMIRNOFF type for Buckingham interactions."""

        _ELEMENT_NAME = "Buckingham"

        a = ParameterAttribute(default=None, unit=unit.kilojoule_per_mole)
        b = ParameterAttribute(default=None, unit=unit.nanometer**-1)
        c = ParameterAttribute(
            default=None,
            unit=unit.kilojoule_per_mole * unit.nanometer**6,
        )

    _TAGNAME = "Buckingham"
    _INFOTYPE = BuckinghamType

    scale12 = ParameterAttribute(default=0.0, converter=float)
    scale13 = ParameterAttribute(default=0.0, converter=float)
    scale14 = ParameterAttribute(default=0.5, converter=float)
    scale15 = ParameterAttribute(default=1.0, converter=float)

    cutoff = ParameterAttribute(default=Quantity("9.0 angstrom"), unit=unit.angstrom)
    switch_width = ParameterAttribute(
        default=Quantity("1.0 angstrom"),
        unit=unit.angstrom,
    )

    periodic_method = ParameterAttribute(
        default="cutoff",
        converter=_allow_only(["cutoff"]),
    )
    nonperiodic_method = ParameterAttribute(
        default="no-cutoff",
        converter=_allow_only(["no-cutoff"]),
    )

    combining_rules = ParameterAttribute(
        default="Lorentz-Berthelot",
        converter=_allow_only(["Lorentz-Berthelot"]),
    )


class DoubleExponentialHandler(ParameterHandler):
    """A custom SMIRNOFF handler for double exponential interactions."""

    class DoubleExponentialType(ParameterType):
        """A custom SMIRNOFF type for double exponential interactions."""

        _ELEMENT_NAME = "Atom"

        r_min = ParameterAttribute(default=None, unit=unit.nanometers)
        epsilon = ParameterAttribute(default=None, unit=unit.kilojoule_per_mole)

    # Give this a different name than the class provided in smirnoff-plugins
    # since the toolkit forbids two handlers from sharing a _TAGNAME
    _TAGNAME = "OtherDoubleExponential"
    _INFOTYPE = DoubleExponentialType

    scale12 = ParameterAttribute(default=0.0, converter=float)
    scale13 = ParameterAttribute(default=0.0, converter=float)
    scale14 = ParameterAttribute(default=0.5, converter=float)
    scale15 = ParameterAttribute(default=1.0, converter=float)

    # These are defined as dimensionless, we should consider enforcing global parameters
    # as being unit-bearing even if that means using `unit.dimensionless`
    alpha = ParameterAttribute(default=18.7)
    beta = ParameterAttribute(default=3.3)

    cutoff = ParameterAttribute(default=9.0 * unit.angstroms, unit=unit.angstrom)
    switch_width = ParameterAttribute(default=1.0 * unit.angstroms, unit=unit.angstrom)

    periodic_method = ParameterAttribute(
        default="cutoff",
        converter=_allow_only(["cutoff"]),
    )
    nonperiodic_method = ParameterAttribute(
        default="no-cutoff",
        converter=_allow_only(["no-cutoff"]),
    )

    combining_rules = ParameterAttribute(
        default="Lorentz-Berthelot",
        converter=_allow_only(["Lorentz-Berthelot"]),
    )


class C4IonHandler(ParameterHandler):
    """A custom SMIRNOFF handler adding C4 interactions from 10.1021/jp505875v."""

    class C4IonType(ParameterType):
        """A custom SMIRNOFF type for C4 ion interactions."""

        _ELEMENT_NAME = "Atom"

        c = ParameterAttribute(
            default=None,
            unit=unit.kilojoule_per_mole * unit.nanometer**4,
        )

    _TAGNAME = "C4Ion"
    _INFOTYPE = C4IonType


class SMIRNOFFBuckinghamCollection(_SMIRNOFFNonbondedCollection):
    """Handler storing vdW potentials as produced by a SMIRNOFF force field."""

    type: Literal["Buckingham"] = "Buckingham"

    is_plugin: bool = True

    acts_as: str = "vdW"

    expression: str = "a*exp(-b*r)-c*r^-6;a=sqrt(a1*a2);b=2/(1/b1+1/b2);c=sqrt(c1*c2);"

    periodic_method: str = "cutoff"
    nonperiodic_method: str = "no-cutoff"

    mixing_rule: str = "Buckingham"

    switch_width: _DistanceQuantity = Quantity(1.0, unit.angstrom)

    @classmethod
    def allowed_parameter_handlers(cls) -> _HandlerIterable:
        """Return a list of allowed types of ParameterHandler classes."""
        return (BuckinghamHandler,)

    @classmethod
    def supported_parameters(cls) -> Iterable[str]:
        """Return a list of supported parameter attributes."""
        return "smirks", "id", "a", "b", "c"

    @classmethod
    def default_parameter_values(cls) -> Iterable[float]:
        """Per-particle parameter values passed to Force.addParticle()."""
        return 0.0, 0.0, 0.0

    @classmethod
    def potential_parameters(cls) -> Iterable[str]:
        """Return a subset of `supported_parameters` that are meant to be included in potentials."""
        return "a", "b", "c"

    @classmethod
    def global_parameters(cls) -> Iterable[str]:
        """Return a list of global parameters, i.e. not per-potential parameters."""
        return tuple()

    def pre_computed_terms(self) -> dict[str, float]:
        """Return a dictionary of pre-computed terms for use in the expression."""
        return dict()

    @classmethod
    def check_openmm_requirements(cls, combine_nonbonded_forces: bool) -> None:
        """Run through a list of assertions about what is compatible when exporting this to OpenMM."""
        assert not combine_nonbonded_forces, (
            "Custom non-bonded functional forms require `combine_nonbonded_forces=False`."
        )

    def store_potentials(self, parameter_handler: BuckinghamHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [TopologyKey, PotentialKey].

        """
        self.periodic_method = parameter_handler.periodic_method.lower()
        self.nonperiodic_method = parameter_handler.nonperiodic_method.lower()
        self.cutoff = parameter_handler.cutoff

        for potential_key in self.key_map.values():
            smirks = potential_key.id
            force_field_parameters = parameter_handler.parameters[smirks]

            potential = Potential(
                parameters={
                    parameter: getattr(force_field_parameters, parameter) for parameter in self.potential_parameters()
                },
            )

            self.potentials[potential_key] = potential

    @classmethod
    def create(  # type: ignore[override]
        cls: _CollectionAlias,
        parameter_handler: BuckinghamHandler,
        topology: Topology,
    ) -> _SMIRNOFFNonbondedCollection:
        """
        Create a SMIRNOFFvdWCollection from toolkit data.

        """
        if type(parameter_handler) not in cls.allowed_parameter_handlers():
            raise InvalidParameterHandlerError(
                f"Found parameter handler type {type(parameter_handler)}, which is not "
                f"supported by potential type {type(cls)}",
            )

        handler = cls(  # type: ignore
            scale_13=parameter_handler.scale13,
            scale_14=parameter_handler.scale14,
            scale_15=parameter_handler.scale15,
            cutoff=parameter_handler.cutoff,
            mixing_rule=parameter_handler.combining_rules.lower(),
            periodic_method=parameter_handler.periodic_method.lower(),
            nonperiodic_method=parameter_handler.nonperiodic_method.lower(),
            switch_width=parameter_handler.switch_width,
        )
        handler.store_matches(parameter_handler=parameter_handler, topology=topology)
        handler.store_potentials(parameter_handler=parameter_handler)

        return handler

    @classmethod
    def parameter_handler_precedence(cls) -> Iterable[str]:
        """
        Return the order in which parameter handlers take precedence when computing charges.
        """
        return "vdw", "VirtualSites"

    def create_virtual_sites(
        self,
        parameter_handler: VirtualSiteHandler,
        topology: Topology,
    ):
        """create() but with virtual sites."""
        raise NotImplementedError()


class SMIRNOFFDoubleExponentialCollection(_SMIRNOFFNonbondedCollection):
    """Handler storing vdW potentials as produced by a SMIRNOFF force field."""

    type: Literal["OtherDoubleExponential"] = "OtherDoubleExponential"

    is_plugin: bool = True

    acts_as: str = "vdW"

    expression: str = (
        "CombinedEpsilon*RepulsionFactor*RepulsionExp-CombinedEpsilon*AttractionFactor*AttractionExp;"
        "CombinedEpsilon=epsilon1*epsilon2;"
        "RepulsionExp=exp(-alpha*ExpDistance);"
        "AttractionExp=exp(-beta*ExpDistance);"
        "ExpDistance=r/CombinedR;"
        "CombinedR=r_min1+r_min2;"
    )

    periodic_method: str = "cutoff"
    nonperiodic_method: str = "no-cutoff"

    mixing_rule: str = ""

    switch_width: _DistanceQuantity = Quantity("1.0 angstrom")

    # TODO: Automagically convert floats into dimensionless Quantity objects
    alpha: _DimensionlessQuantity
    beta: _DimensionlessQuantity

    @classmethod
    def allowed_parameter_handlers(cls) -> _HandlerIterable:
        """Return a list of allowed types of ParameterHandler classes."""
        return (DoubleExponentialHandler,)

    @classmethod
    def supported_parameters(cls) -> Iterable[str]:
        """Return a list of supported parameter attributes."""
        return "smirks", "id", "r_min", "epsilon"

    @classmethod
    def default_parameter_values(cls) -> Iterable[float]:
        """Per-particle parameter values passed to Force.addParticle()."""
        return 0.0, 0.0

    @classmethod
    def potential_parameters(cls) -> Iterable[str]:
        """Return a subset of `supported_parameters` that are meant to be included in potentials."""
        return "r_min", "epsilon"

    @classmethod
    def global_parameters(cls) -> Iterable[str]:
        """Return a list of global parameters, i.e. not per-potential parameters."""
        return "alpha", "beta"

    def pre_computed_terms(self) -> dict[str, float]:
        """Return a dictionary of pre-computed terms for use in the expression."""
        alpha_min_beta = self.alpha - self.beta

        return {
            "AlphaMinBeta": alpha_min_beta,
            "RepulsionFactor": self.beta * math.exp(self.alpha) / alpha_min_beta,
            "AttractionFactor": self.alpha * math.exp(self.beta) / alpha_min_beta,
        }

    def modify_parameters(
        self,
        original_parameters: dict[str, Quantity],
    ) -> dict[str, float]:
        """Optionally modify parameters prior to their being stored in a force."""
        # It's important that these keys are in the order of self.potential_parameters(),
        # consider adding a check somewhere that this is the case.
        _units = {"r_min": unit.nanometer, "epsilon": unit.kilojoule_per_mole}
        return {
            "r_min": original_parameters["r_min"].m_as(_units["r_min"]) * 0.5,
            "epsilon": math.sqrt(
                original_parameters["epsilon"].m_as(_units["epsilon"]),
            ),
        }

    @classmethod
    def check_openmm_requirements(cls, combine_nonbonded_forces: bool) -> None:
        """Run through a list of assertions about what is compatible when exporting this to OpenMM."""
        assert not combine_nonbonded_forces, (
            "Custom non-bonded functional forms require `combine_nonbonded_forces=False`."
        )

    def store_potentials(self, parameter_handler: DoubleExponentialHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [TopologyKey, PotentialKey].

        """
        self.periodic_method = parameter_handler.periodic_method.lower()
        self.nonperiodic_method = parameter_handler.nonperiodic_method.lower()
        self.cutoff = parameter_handler.cutoff

        for potential_key in self.key_map.values():
            smirks = potential_key.id
            force_field_parameters = parameter_handler.parameters[smirks]

            potential = Potential(
                parameters={
                    parameter: getattr(force_field_parameters, parameter) for parameter in self.potential_parameters()
                },
            )

            self.potentials[potential_key] = potential

    @classmethod
    def create(  # type: ignore[override]
        cls: _CollectionAlias,
        parameter_handler: DoubleExponentialHandler,
        topology: Topology,
    ) -> _SMIRNOFFNonbondedCollection:
        """
        Create a SMIRNOFFvdWCollection from toolkit data.

        """
        if type(parameter_handler) not in cls.allowed_parameter_handlers():
            raise InvalidParameterHandlerError(
                f"Found parameter handler type {type(parameter_handler)}, which is not "
                f"supported by potential type {type(cls)}",
            )

        handler = cls(  # type: ignore
            alpha=Quantity(parameter_handler.alpha),
            beta=Quantity(parameter_handler.beta),
            scale_13=parameter_handler.scale13,
            scale_14=parameter_handler.scale14,
            scale_15=parameter_handler.scale15,
            cutoff=parameter_handler.cutoff,
            mixing_rule=parameter_handler.combining_rules.lower(),
            periodic_method=parameter_handler.periodic_method.lower(),
            nonperiodic_method=parameter_handler.nonperiodic_method.lower(),
            switch_width=parameter_handler.switch_width,
        )
        handler.store_matches(parameter_handler=parameter_handler, topology=topology)
        handler.store_potentials(parameter_handler=parameter_handler)

        return handler

    @classmethod
    def parameter_handler_precedence(cls) -> Iterable[str]:
        """
        Return the order in which parameter handlers take precedence when computing charges.
        """
        return "vdw", "VirtualSites"

    def create_virtual_sites(
        self,
        parameter_handler: VirtualSiteHandler,
        topology: Topology,
    ):
        """create() but with virtual sites."""
        raise NotImplementedError()


class SMIRNOFFC4IonCollection(_SMIRNOFFNonbondedCollection):
    """Handler storing vdW potentials as produced by a SMIRNOFF force field."""

    type: Literal["C4Ion"] = "C4Ion"

    is_plugin: bool = True

    expression: str = "c*r^-4;c=sqrt(c1*c2);"

    periodic_method: str = "cutoff"
    nonperiodic_method: str = "no-cutoff"

    @classmethod
    def allowed_parameter_handlers(cls) -> _HandlerIterable:
        """Return a list of allowed types of ParameterHandler classes."""
        return (C4IonHandler,)

    @classmethod
    def supported_parameters(cls) -> Iterable[str]:
        """Return a list of supported parameter attributes."""
        return "smirks", "id", "c"

    @classmethod
    def default_parameter_values(cls) -> Iterable[float]:
        """Per-particle parameter values passed to Force.addParticle()."""
        return (0.0,)

    @classmethod
    def potential_parameters(cls) -> Iterable[str]:
        """Return a subset of `supported_parameters` that are meant to be included in potentials."""
        return "c"

    @classmethod
    def check_openmm_requirements(cls, combine_nonbonded_forces: bool) -> None:
        """Run through a list of assertions about what is compatible when exporting this to OpenMM."""
        assert combine_nonbonded_forces, "The r ** -4 term is only implemented with a single `NonbondedForce`."

    def store_potentials(self, parameter_handler: DoubleExponentialHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [TopologyKey, PotentialKey].

        """
        for potential_key in self.key_map.values():
            smirks = potential_key.id
            force_field_parameters = parameter_handler.parameters[smirks]

            potential = Potential(
                parameters={
                    parameter: getattr(force_field_parameters, parameter) for parameter in self.potential_parameters()
                },
            )

            self.potentials[potential_key] = potential

    def modify_openmm_forces(
        self,
        interchange,
        system,
        add_constrained_forces,
        constrained_pairs,
        particle_map,
    ):
        """Add a `openmm.CustomNonbondedForce` to handle the C/r^4 term. See 10.1021/jp505875v."""
        import openmm
        from openff.interchange.models import TopologyKey

        non_bonded_forces = [
            force
            for force in system.getForces()
            if isinstance(force, openmm.NonbondedForce | openmm.CustomNonbondedForce)
        ]

        if len(non_bonded_forces) != 1:
            raise NotImplementedError(
                "This handler is only compatible with a single NonbondedForce/CustomNonbondedForce.",
            )

        if type(non_bonded_forces[0]) is openmm.CustomNonbondedForce:
            raise NotImplementedError(
                "This handler is only compatible with a single NonbondedForce/CustomNonbondedForce.",
            )

        non_bonded_force = non_bonded_forces[0]

        c4_force = openmm.CustomNonbondedForce(self.expression)

        system.addForce(c4_force)

        c4_force.addPerParticleParameter("c")

        for molecule in interchange.topology.molecules:
            for atom in molecule.atoms:
                atom_index = interchange.topology.atom_index(atom)

                top_key = TopologyKey(atom_indices=(atom_index,))

                if top_key not in self.potentials:
                    # This handler only adds interactions to highly-charged ions, so many
                    # atom indices will not be found in the key map. OpenMM requires all
                    # forces have as many particles as are found in the system, so just add
                    # this particle with a zeroed (numerator) parameters.
                    c4_force.addParticle([0.0])

                else:
                    c4_force.addParticle(
                        [self.potentials[top_key].parameters["c"]],
                    )

        # OpenMM requires "All Forces must have identical exclusions", so just copy them over
        for exception_index in range(non_bonded_force.getNumExceptions()):
            p1, p2, *_ = non_bonded_force.getExceptionParameters(exception_index)
            c4_force.addExclusion(p1, p2)

        # Copy non-bonded settings from the main `NonbondedForce`
        c4_force.setNonbondedMethod(non_bonded_force.getNonbondedMethod())
        c4_force.setCutoffDistance(non_bonded_force.getCutoffDistance())
        c4_force.setUseSwitchingFunction(non_bonded_force.getUseSwitchingFunction())
        c4_force.setSwitchingDistance(non_bonded_force.getSwitchingDistance())
        c4_force.setUseLongRangeCorrection(
            non_bonded_force.getUseDispersionCorrection(),
        )



from pathlib import Path


def make_monolithic(topology_file: str, keep_file: bool = False) -> list[str]:
    parent = Path(topology_file).parent
    name = Path(topology_file).name

    monolithic_path = parent / f"MONOLITHIC_{name}"

    return_value = list()

    with open(topology_file) as in_obj, open(monolithic_path, "w") as out_obj:
        for line in in_obj:
            if len(line) == 0:
                continue

            stripped = line.strip().split()

            if len(stripped) == 0:
                continue

            if stripped[0].startswith(";"):
                continue

            if stripped[0].startswith("#include"):
                itp_file = stripped[1].strip('"')

                with open(itp_file) as itp_obj:
                    for itp_line in itp_obj:
                        if len(itp_line) == 0:
                            continue

                        if itp_line.startswith(";"):
                            continue

                        # TODO: itp files can load from itp files, this block should recurse

                        return_value.append(itp_line)
                        out_obj.write(itp_line)

            else:
                return_value.append(line)
                out_obj.write(line)

    if Path(monolithic_path).exists() and not keep_file:
        Path(monolithic_path).unlink()

    return return_value



import os.path
import tempfile
import time

import polars
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from tqdm import tqdm

from openff.interchange import __version__

data_frame_file = f"{__version__}/ligand_in_water.csv"

if os.path.exists(data_frame_file):
    exit("CSV file already exists, exiting")

water = Molecule.from_smiles("O")
paracetamol = Molecule.from_smiles("CC(=O)NC1=CC=C(C=C1)O")

water.generate_conformers(n_conformers=1)
paracetamol.generate_conformers(n_conformers=1)

interchange = ForceField("openff-2.3.0-rc1.offxml").create_interchange(
    Topology.from_molecules(
        [paracetamol] + 10000 * [water],
    ),
)

interchange.box = Quantity([4, 4, 4], "nanometer")

times = list()

with tempfile.TemporaryDirectory() as tmpdir:
    for index in tqdm(range(10), desc="Ligand in water benchmark"):
        openmm_start = time.perf_counter()
        interchange.to_openmm()
        openmm_end = time.perf_counter()

        gromacs_start = time.perf_counter()
        interchange.to_gromacs(prefix=f"{tmpdir}/temp")
        gromacs_end = time.perf_counter()

        times.append(
            {
                "iteration": index,
                "openmm_time": openmm_end - openmm_start,
                "gromacs_time": gromacs_end - gromacs_start,
            },
        )

data_frame = polars.from_dicts(
    data=times,
    schema={
        "iteration": polars.Int64,
        "openmm_time": polars.Float64,
        "gromacs_time": polars.Float64,
    },
)

data_frame.write_csv(data_frame_file)



from importlib import resources

import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff import VirtualSiteHandler
from openff.units import ensure_quantity
from openff.utilities import get_data_file_path, has_package, skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer, get_protein, needs_gmx
from openff.interchange.drivers import get_gromacs_energies, get_openmm_energies
from openff.interchange.exceptions import (
    UnsupportedExportError,
    VirtualSiteTypeNotImplementedError,
)
from openff.interchange.interop.gromacs._import._import import (
    _read_box,
    _read_coordinates,
)

if has_package("openmm"):
    import openmm.app
    import openmm.unit


@skip_if_missing("intermol")
@skip_if_missing("mdtraj")
@skip_if_missing("openmm")
@needs_gmx
class TestGROMACSGROFile:
    try:
        _INTERMOL_PATH = resources.files(
            "intermol.tests.gromacs.unit_tests",
        )
    except ImportError:
        _INTERMOL_PATH = None

    @skip_if_missing("intermol")
    def test_load_gro(self):
        file = self._INTERMOL_PATH / "angle10_vacuum/angle10_vacuum.gro"

        positions = _read_coordinates(file)
        box = _read_box(file)

        openmm_gro = openmm.app.GromacsGroFile(file)
        openmm_positions = ensure_quantity(
            openmm_gro.getPositions(),
            "openff",
        )
        openmm_box = ensure_quantity(
            openmm_gro.getPeriodicBoxVectors(),
            "openff",
        )

        assert numpy.allclose(positions, openmm_positions)
        assert numpy.allclose(box, openmm_box)

    @skip_if_missing("intermol")
    @pytest.mark.skip("don't run parmed tests")
    def test_load_gro_nonstandard_precision(self):
        file = self._INTERMOL_PATH / "lj3_bulk/lj3_bulk.gro"

        coords = _read_coordinates(file).m_as("nanometer")

        # OpenMM seems to assume a precision of 3. Use InterMol instead here.
        from intermol.gromacs.grofile_parser import GromacsGroParser

        intermol_gro = GromacsGroParser(file)
        intermol_gro.read()

        # InterMol stores positions as NumPy arrays _of openmm.unit.Quantity_
        # objects, so need to carefully convert everything through unitless
        # floats; converting through Pint-like quantities might produce a
        # funky double-wrapped thing
        def converter(x):
            return x.value_in_unit(openmm.unit.nanometer)

        other_coords = numpy.frompyfunc(converter, 1, 1)(intermol_gro.positions).astype(
            float,
        )

        assert numpy.allclose(coords, other_coords)

        # This file happens to have 12 digits of preicion; what really matters is that
        # the convential precision of 3 was not used.
        n_decimals = len(str(coords[0, 0]).split(".")[1])
        assert n_decimals == 12


@needs_gmx
class TestGROMACS:
    @pytest.mark.slow
    @pytest.mark.skip("from_top is not yet refactored for new Topology API")
    @pytest.mark.parametrize("reader", ["intermol", "internal"])
    @pytest.mark.parametrize(
        "smiles",
        [
            "C",
            "O=C=O",  # Adds unconstrained bonds without torsion(s)
            "CC",  # Adds a proper torsion term(s)
            # "C=O",  # Simplest molecule with any improper torsion
            "OC=O",  # Simplest molecule with a multi-term torsion
            # "CCOC",  # This hits t86, which has a non-1.0 idivf
            # "C1COC(=O)O1",  # This adds an improper, i2
        ],
    )
    def test_simple_roundtrip(self, sage, smiles, reader):
        molecule = MoleculeWithConformer.from_smiles(smiles)
        molecule.name = molecule.to_hill_formula()
        topology = molecule.to_topology()

        out = Interchange.from_smirnoff(force_field=sage, topology=topology)
        out.box = [4, 4, 4]
        out.positions = molecule.conformers[0]

        out.to_top("out.top")
        out.to_gro("out.gro")

        converted = Interchange.from_gromacs("out.top", "out.gro", reader=reader)

        assert numpy.allclose(out.positions, converted.positions)
        assert numpy.allclose(out.box, converted.box)

        get_gromacs_energies(out).compare(
            get_gromacs_energies(converted),
            tolerances={
                "Bond": Quantity(0.002 * molecule.n_bonds, "kilojoule / mol"),
                "Electrostatics": Quantity(0.05, "kilojoule / mol"),
            },
        )

    @pytest.mark.skip("don't run parmed tests")
    def test_num_impropers(self, sage):
        parmed = pytest.importorskip("parmed")
        out = Interchange.from_smirnoff(
            sage,
            MoleculeWithConformer.from_smiles("CC1=CC=CC=C1").to_topology(),
        )

        out.box = Quantity(4 * numpy.eye(3), "nanometer")
        out.to_top("tmp.top")

        # Sanity check; toluene should have some improper(s)
        assert len(out["ImproperTorsions"].key_map) > 0

        struct = parmed.load_file("tmp.top")
        n_impropers_parmed = len([d for d in struct.dihedrals if d.improper])
        assert n_impropers_parmed == len(out["ImproperTorsions"].key_map)

    @pytest.mark.slow
    @skip_if_missing("intermol")
    @pytest.mark.skip(reason="Re-implement when SMIRNOFF supports more mixing rules")
    def test_set_mixing_rule(self, ethanol_top, sage):
        from intermol.gromacs.gromacs_parser import GromacsParser

        interchange = Interchange.from_smirnoff(force_field=sage, topology=ethanol_top)
        interchange.positions = numpy.zeros((ethanol_top.n_atoms, 3))
        interchange.to_gro("tmp.gro")

        interchange.box = [4, 4, 4]
        interchange.to_top("lorentz.top")
        lorentz = GromacsParser("lorentz.top", "tmp.gro").read()
        assert lorentz.combination_rule == "Lorentz-Berthelot"

        interchange["vdW"].mixing_rule = "geometric"

        interchange.to_top("geometric.top")
        geometric = GromacsParser("geometric.top", "tmp.gro").read()
        assert geometric.combination_rule == "Multiply-Sigeps"

    @pytest.mark.skip(reason="Re-implement when SMIRNOFF supports more mixing rules")
    def test_unsupported_mixing_rule(self, ethanol_top, sage):
        # TODO: Update this test when the model supports more mixing rules than GROMACS does
        interchange = Interchange.from_smirnoff(force_field=sage, topology=ethanol_top)
        interchange["vdW"].mixing_rule = "kong"

        with pytest.raises(UnsupportedExportError, match="rule `geometric` not compat"):
            interchange.to_top("out.top")

    @pytest.mark.slow
    @pytest.mark.skip("don't run parmed tests")
    @skip_if_missing("openmm")
    def test_residue_info(self, sage):
        """Test that residue information is passed through to .top files."""
        from openff.units.openmm import from_openmm

        parmed = pytest.importorskip("parmed")

        protein = get_protein("MainChain_HIE")

        box_vectors = from_openmm(
            openmm.app.PDBFile(
                get_data_file_path(
                    "proteins/MainChain_HIE.pdb",
                    "openff.toolkit",
                ),
            ).topology.getPeriodicBoxVectors(),
        )

        ff14sb = ForceField("ff14sb_off_impropers_0.0.3.offxml")

        out = Interchange.from_smirnoff(
            force_field=ff14sb,
            topology=[protein],
            box=box_vectors,
        )

        out.to_top("tmp.top")

        parmed_structure = parmed.load_file("tmp.top")

        for found_residue, original_residue in zip(
            parmed_structure.residues,
            out.topology.hierarchy_iterator("residues"),
        ):
            assert found_residue.name == original_residue.residue_name
            assert str(found_residue.number + 1) == original_residue.residue_number

    @pytest.mark.skip("Broken, unclear if cases like these are worth supporting")
    def test_nonconsecutive_isomorphic_molecules(self, sage_unconstrained):
        molecules = [Molecule.from_smiles(smiles) for smiles in ["CC", "CCO", "CC"]]

        for index, molecule in enumerate(molecules):
            molecule.generate_conformers(n_conformers=1)
            molecule.conformers[0] += Quantity(3 * [5 * index], "angstrom")

        topology = Topology.from_molecules(molecules)
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        out = Interchange.from_smirnoff(sage_unconstrained, topology)

        get_gromacs_energies(out).compare(
            get_openmm_energies(out, combine_nonbonded_forces=True),
            {"Nonbonded": Quantity(0.5, "kilojoule_per_mole")},
        )


@needs_gmx
@pytest.mark.skip("Needs rewrite")
class TestGROMACSVirtualSites:
    @pytest.fixture
    def sigma_hole_type(self, sage):
        """A handler with a bond charge virtual site on a C-Cl bond."""
        return VirtualSiteHandler.VirtualSiteBondChargeType(
            name="EP",
            smirks="[#6:1]-[#17:2]",
            distance=Quantity(1.4, "angstrom"),
            type="BondCharge",
            match="once",
            charge_increment1=Quantity(0.1, "elementary_charge"),
            charge_increment2=Quantity(0.2, "elementary_charge"),
        )

    @pytest.fixture
    def sage_with_monovalent_lone_pair(self, sage):
        """Fixture that loads an SMIRNOFF XML for argon"""
        virtual_site_handler = VirtualSiteHandler(version=0.3)

        carbonyl_type = VirtualSiteHandler.VirtualSiteType(
            name="EP",
            smirks="[O:1]=[C:2]-[*:3]",
            distance=Quantity(0.3, "angstrom"),
            type="MonovalentLonePair",
            match="all_permutations",
            outOfPlaneAngle=Quantity(0.0, "degree"),
            inPlaneAngle=Quantity(120.0, "degree"),
            charge_increment1=Quantity(0.05, "elementary_charge"),
            charge_increment2=Quantity(0.1, "elementary_charge"),
            charge_increment3=Quantity(0.15, "elementary_charge"),
        )

        virtual_site_handler.add_parameter(parameter=carbonyl_type)
        sage.register_parameter_handler(virtual_site_handler)

        return sage

    @pytest.mark.xfail
    @skip_if_missing("parmed")
    def test_sigma_hole_example(self, sage_with_sigma_hole):
        """Test that a single-molecule sigma hole example runs"""
        parmed = pytest.importorskip("parmed")

        molecule = MoleculeWithConformer.from_smiles("CCl", name="Chloromethane")

        out = Interchange.from_smirnoff(
            force_field=sage_with_sigma_hole,
            topology=molecule.to_topology(),
        )
        out.box = [4, 4, 4]
        out.positions = molecule.conformers[0]

        # TODO: Sanity-check reported energies
        get_gromacs_energies(out)

        out.to_top("sigma.top")
        gmx_top = parmed.load_file("sigma.top")

        assert abs(numpy.sum([p.charge for p in gmx_top.atoms])) < 1e-3

    def test_carbonyl_example(self, sage_with_monovalent_lone_pair):
        """Test that a single-molecule DivalentLonePair example runs"""
        mol = MoleculeWithConformer.from_smiles("C=O", name="Carbon_monoxide")

        out = Interchange.from_smirnoff(
            force_field=sage_with_monovalent_lone_pair,
            topology=mol.to_topology(),
        )
        out.box = [4, 4, 4]
        out.positions = mol.conformers[0]

        with pytest.raises(
            VirtualSiteTypeNotImplementedError,
            match=r"MonovalentLonePair not implemented.",
        ):
            # TODO: Sanity-check reported energies
            get_gromacs_energies(out)



import json

from openff.toolkit import Quantity
from openff.utilities.testing import skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer, needs_gmx
from openff.interchange.drivers import get_gromacs_energies, get_openmm_energies


@needs_gmx
@skip_if_missing("openmm")
def test_issue_908(sage_unconstrained):
    molecule = MoleculeWithConformer.from_smiles("ClCCl")
    topology = molecule.to_topology()
    topology.box_vectors = Quantity([4, 4, 4], "nanometer")

    state1 = sage_unconstrained.create_interchange(topology)

    with open("test.json", "w") as f:
        f.write(state1.model_dump_json())

    state2 = Interchange.model_validate(
        json.load(
            open("test.json"),
        ),
    )

    assert state2["Electrostatics"].scale_14 == 0.8333333333

    get_gromacs_energies(state1).compare(get_gromacs_energies(state2))
    get_openmm_energies(
        state1,
        combine_nonbonded_forces=False,
    ).compare(get_openmm_energies(state2, combine_nonbonded_forces=False))



"""
Helper functions for producing `openmm.Force` objects for valence terms.
"""

import functools

from openff.units.openmm import to_openmm as to_openmm_quantity
from openff.utilities.utilities import has_package

from openff.interchange.exceptions import UnsupportedExportError
from openff.interchange.models import BaseVirtualSiteKey, PotentialKey
from openff.interchange.smirnoff._valence import (
    SMIRNOFFAngleCollection,
    SMIRNOFFBondCollection,
    SMIRNOFFImproperTorsionCollection,
    SMIRNOFFProperTorsionCollection,
)

if has_package("openmm"):
    import openmm


def _process_constraints(
    interchange,
    openmm_sys,
    particle_map: dict[int | BaseVirtualSiteKey, int],
) -> set[tuple[int, ...]]:
    """
    Process the Constraints section of an Interchange object.
    """
    try:
        constraint_handler = interchange["Constraints"]
    except LookupError:
        return set()

    constrained_pairs: set[tuple[int, ...]] = set()

    for top_key, pot_key in constraint_handler.key_map.items():
        openff_indices = top_key.atom_indices

        openmm_indices = tuple(particle_map[index] for index in openff_indices)

        params = constraint_handler.potentials[pot_key].parameters
        distance = params["distance"]
        distance_omm = distance.m_as("nanometer")

        constrained_pairs.add(tuple(sorted(openmm_indices)))
        openmm_sys.addConstraint(
            openmm_indices[0],
            openmm_indices[1],
            distance_omm,
        )

    return constrained_pairs


@functools.lru_cache
def _lookup_bond_parameters_from_potential_key(
    collection: SMIRNOFFBondCollection,
    potential_key: PotentialKey,
) -> tuple[float, float]:
    parameters = collection.potentials[potential_key].parameters

    return (
        parameters["k"].m_as("kilojoule / nanometer ** 2 / mole"),
        parameters["length"].m_as("nanometer"),
    )


def _process_bond_forces(
    interchange,
    openmm_sys,
    add_constrained_forces: bool,
    constrained_pairs: set[tuple[int, ...]],
    particle_map: dict[int | BaseVirtualSiteKey, int],
):
    """
    Process the Bonds section of an Interchange object.
    """
    _lookup_bond_parameters_from_potential_key.cache_clear()

    try:
        bond_handler = interchange.collections["Bonds"]
    except KeyError:
        return

    if bond_handler.expression == "k/2*(r-length)**2":
        harmonic_bond_force = openmm.HarmonicBondForce()
    else:
        raise UnsupportedExportError(
            f"Found an unsupported functional form in the bond handler:\n\t{bond_handler.expression=}",
        )

    openmm_sys.addForce(harmonic_bond_force)

    has_constraint_handler = "Constraints" in interchange.collections

    for top_key, pot_key in bond_handler.key_map.items():
        openff_indices = top_key.atom_indices
        openmm_indices = tuple(particle_map[index] for index in openff_indices)

        if has_constraint_handler and not add_constrained_forces:
            if _is_constrained(
                constrained_pairs,
                (openmm_indices[0], openmm_indices[1]),
            ):
                # This bond's length is constrained, dpo so not add a bond force
                continue

        k, length = _lookup_bond_parameters_from_potential_key(bond_handler, pot_key)

        harmonic_bond_force.addBond(
            particle1=openmm_indices[0],
            particle2=openmm_indices[1],
            length=length,
            k=k,
        )


@functools.lru_cache
def _lookup_angle_parameters_from_potential_key(
    collection: SMIRNOFFAngleCollection,
    potential_key: PotentialKey,
) -> tuple[float, float]:
    parameters = collection.potentials[potential_key].parameters

    return (
        parameters["k"].m_as("kilojoule / radian / mole"),
        parameters["angle"].m_as("radian"),
    )


def _process_angle_forces(
    interchange,
    openmm_sys,
    add_constrained_forces: bool,
    constrained_pairs: set[tuple[int, ...]],
    particle_map,
):
    """
    Process the Angles section of an Interchange object.
    """
    _lookup_angle_parameters_from_potential_key.cache_clear()

    try:
        angle_handler = interchange.collections["Angles"]
    except KeyError:
        return

    if angle_handler.expression == "k/2*(theta-angle)**2":
        custom = False
        harmonic_angle_force = openmm.HarmonicAngleForce()
    elif angle_handler.expression == "k/2*(cos(theta)-cos(angle))**2":
        custom = True
        harmonic_angle_force = openmm.CustomAngleForce(
            angle_handler.expression.replace("**", "^"),
        )
        for parameter_name in angle_handler.potential_parameters():
            harmonic_angle_force.addPerAngleParameter(parameter_name)
    else:
        raise UnsupportedExportError(
            f"Found an unsupported functional form in the angle handler:\n\t{angle_handler.expression=}",
        )

    openmm_sys.addForce(harmonic_angle_force)

    has_constraint_handler = "Constraints" in interchange.collections

    for top_key, pot_key in angle_handler.key_map.items():
        openff_indices = top_key.atom_indices
        openmm_indices = tuple(particle_map[index] for index in openff_indices)

        if has_constraint_handler and not add_constrained_forces:
            if _is_constrained(
                constrained_pairs,
                (openmm_indices[0], openmm_indices[2]),
            ):
                if _is_constrained(
                    constrained_pairs,
                    (openmm_indices[0], openmm_indices[1]),
                ):
                    if _is_constrained(
                        constrained_pairs,
                        (openmm_indices[1], openmm_indices[2]),
                    ):
                        # This angle's geometry is fully subject to constraints, so do
                        # not an angle force
                        continue

        if custom:
            params = angle_handler.potentials[pot_key].parameters
            parameter_values = [to_openmm_quantity(params[val]) for val in angle_handler.potential_parameters()]

            harmonic_angle_force.addAngle(
                openmm_indices[0],
                openmm_indices[1],
                openmm_indices[2],
                parameter_values,
            )

        else:
            k, angle = _lookup_angle_parameters_from_potential_key(angle_handler, pot_key)

            harmonic_angle_force.addAngle(
                particle1=openmm_indices[0],
                particle2=openmm_indices[1],
                particle3=openmm_indices[2],
                angle=angle,
                k=k,
            )


def _process_torsion_forces(interchange, openmm_sys, particle_map):
    if "ProperTorsions" in interchange.collections:
        _process_proper_torsion_forces(interchange, openmm_sys, particle_map)
    if "RBTorsions" in interchange.collections:
        _process_rb_torsion_forces(interchange, openmm_sys, particle_map)


@functools.lru_cache
def _lookup_torsion_parameters_from_potential_key(
    collection: SMIRNOFFProperTorsionCollection | SMIRNOFFImproperTorsionCollection,
    potential_key: PotentialKey,
) -> tuple[float, int, float, int]:
    parameters = collection.potentials[potential_key].parameters

    # float rounding can sometimes make Quantity(1) into 0.9999999999999999
    return (
        parameters["k"].m_as("kilojoule / mole"),
        int(round(parameters["periodicity"], 10)),
        parameters["phase"].m_as("radian"),
        int(round(parameters["idivf"], 10)),
    )


def _process_proper_torsion_forces(interchange, openmm_sys, particle_map):
    """
    Process the Propers section of an Interchange object.
    """
    _lookup_torsion_parameters_from_potential_key.cache_clear()

    torsion_force = openmm.PeriodicTorsionForce()
    openmm_sys.addForce(torsion_force)

    proper_torsion_handler = interchange["ProperTorsions"]

    for top_key, pot_key in proper_torsion_handler.key_map.items():
        openff_indices = top_key.atom_indices
        openmm_indices = tuple(particle_map[index] for index in openff_indices)

        k, periodicity, phase, idivf = _lookup_torsion_parameters_from_potential_key(proper_torsion_handler, pot_key)

        if idivf == 0:
            raise RuntimeError("Found an idivf of 0.")

        torsion_force.addTorsion(
            openmm_indices[0],
            openmm_indices[1],
            openmm_indices[2],
            openmm_indices[3],
            periodicity,
            phase,
            k / idivf,
        )


def _process_rb_torsion_forces(interchange, openmm_sys, particle_map):
    """
    Process Ryckaert-Bellemans torsions.
    """
    rb_force = openmm.RBTorsionForce()
    openmm_sys.addForce(rb_force)

    rb_torsion_handler = interchange["RBTorsions"]

    for top_key, pot_key in rb_torsion_handler.key_map.items():
        openff_indices = top_key.atom_indices
        openmm_indices = tuple(particle_map[index] for index in openff_indices)

        params = rb_torsion_handler.potentials[pot_key].parameters

        c0 = params["c0"].m_as("kilojoule / mole")
        c1 = params["c1"].m_as("kilojoule / mole")
        c2 = params["c2"].m_as("kilojoule / mole")
        c3 = params["c3"].m_as("kilojoule / mole")
        c4 = params["c4"].m_as("kilojoule / mole")
        c5 = params["c5"].m_as("kilojoule / mole")

        rb_force.addTorsion(
            openmm_indices[0],
            openmm_indices[1],
            openmm_indices[2],
            openmm_indices[3],
            c0,
            c1,
            c2,
            c3,
            c4,
            c5,
        )


def _process_improper_torsion_forces(interchange, openmm_sys, particle_map):
    """
    Process the Impropers section of an Interchange object.
    """
    # Possible performance improvement to be had by more carefully managing this cache, see:
    # https://github.com/openforcefield/openff-interchange/pull/1452#discussion_r2891781038
    _lookup_torsion_parameters_from_potential_key.cache_clear()

    if "ImproperTorsions" not in interchange.collections.keys():
        return

    for force in openmm_sys.getForces():
        if type(force) is openmm.PeriodicTorsionForce:
            torsion_force = force
            break
    else:
        torsion_force = openmm.PeriodicTorsionForce()
        openmm_sys.addForce(torsion_force)

    improper_torsion_handler = interchange["ImproperTorsions"]

    for top_key, pot_key in improper_torsion_handler.key_map.items():
        openff_indices = top_key.atom_indices
        openmm_indices = tuple(particle_map[index] for index in openff_indices)

        k, periodicity, phase, idivf = _lookup_torsion_parameters_from_potential_key(improper_torsion_handler, pot_key)

        torsion_force.addTorsion(
            openmm_indices[0],
            openmm_indices[1],
            openmm_indices[2],
            openmm_indices[3],
            periodicity,
            phase,
            k / idivf,
        )


def _is_constrained(
    constrained_pairs: set[tuple[int, ...]],
    pair: tuple[int, int],
) -> bool:
    if (pair[0], pair[1]) in constrained_pairs:
        return True
    if (pair[1], pair[0]) in constrained_pairs:
        return True
    return False



import uuid
from io import StringIO

from openff.toolkit.utils._viz import TopologyNGLViewStructure

from openff.interchange import Interchange


class InterchangeNGLViewStructure(TopologyNGLViewStructure):
    """Subclass of the toolkit's NGLView interface."""

    def __init__(
        self,
        interchange: Interchange,
        ext: str = "PDB",
    ):
        self.interchange = interchange
        self.ext = ext.lower()
        self.params: dict = dict()
        self.id = str(uuid.uuid4())

    def get_structure_string(self) -> str:
        """Get structure as a string."""
        with StringIO() as f:
            self.interchange.to_pdb(f, include_virtual_sites=True)
            structure_string = f.getvalue()
        return structure_string



"""The logic behind `Interchange.combine`."""

import copy
import logging
import warnings
from typing import TYPE_CHECKING

import numpy
from openff.toolkit import Quantity

from openff.interchange.components.toolkit import _combine_topologies
from openff.interchange.exceptions import (
    CutoffMismatchError,
    SwitchingFunctionMismatchError,
    UnsupportedCombinationError,
)
from openff.interchange.warnings import InterchangeCombinationWarning

if TYPE_CHECKING:
    from openff.interchange import Interchange


logger = logging.getLogger(__name__)

DEFAULT_CUTOFF_TOLERANCE = Quantity("1e-6 nanometer")


def _are_almost_equal(
    target: Quantity | float,
    reference: Quantity | float,
    tolerance: Quantity | float,
):
    return abs(reference - target) < tolerance


def _check_nonbonded_compatibility(
    interchange1: "Interchange",
    interchange2: "Interchange",
):
    if not (
        "vdW" in interchange1.collections
        and "vdW" in interchange2.collections
        and "Electrostatics" in interchange1.collections
        and "Electrostatics" in interchange2.collections
    ):
        raise UnsupportedCombinationError(
            "One or more inputs is missing a vdW and/or Electrostatics collection(s).",
        )

    for key in ["vdW", "Electrostatics"]:
        if not _are_almost_equal(
            interchange1[key].cutoff,  # type: ignore[attr-defined]
            interchange2[key].cutoff,  # type: ignore[attr-defined]
            DEFAULT_CUTOFF_TOLERANCE,
        ):
            raise CutoffMismatchError(
                f"{key} cutoffs do not match. Found {interchange1[key].cutoff} and {interchange2[key].cutoff}.",  # type: ignore[attr-defined]
            )

    if not _are_almost_equal(
        interchange1["vdW"].switch_width,
        interchange2["vdW"].switch_width,
        DEFAULT_CUTOFF_TOLERANCE,
    ):
        raise SwitchingFunctionMismatchError(
            f"Switching distance(s) do not match. Found "
            f"{interchange1['vdW'].switch_width} and {interchange2['vdW'].switch_width}.",
        )

    if not _are_almost_equal(
        interchange1["vdW"].scale_14,
        interchange2["vdW"].scale_14,
        1e-6,
    ):
        raise UnsupportedCombinationError(
            "1-4 scaling factors in vdW handler(s) do not match.",
        )

    if interchange1["Electrostatics"].scale_14 != interchange2["Electrostatics"].scale_14:
        if sorted([interchange1["Electrostatics"].scale_14, interchange2["Electrostatics"].scale_14]) == [
            0.833333,
            0.8333333333,
        ]:
            warnings.warn(
                "Found electrostatics 1-4 scaling factors of 5/6 with slightly different rounding "
                "(0.833333 and 0.8333333333). This likely stems from OpenFF using more digits in rounding 1/1.2. "
                "The value of 0.8333333333 will be used, which may or may not introduce small errors. ",
                InterchangeCombinationWarning,
            )

            interchange1["Electrostatics"].scale_14 = 0.8333333333
            interchange2["Electrostatics"].scale_14 = 0.8333333333

        else:
            raise UnsupportedCombinationError(
                "1-4 scaling factors in Electrostatics handler(s) do not match.",
            )


def _combine(
    input1: "Interchange",
    input2: "Interchange",
) -> "Interchange":
    warnings.warn(
        "Interchange object combination is complex and may produce strange results outside "
        "of use cases it has been tested in. Use with caution and thoroughly validate results!",
        InterchangeCombinationWarning,
    )

    result = copy.deepcopy(input1)

    result.topology = _combine_topologies(input1.topology, input2.topology)
    atom_offset = input1.topology.n_atoms

    _check_nonbonded_compatibility(input1, input2)

    for collection in input2.collections.values():
        for potential_key in collection.potentials:
            if "_DUPLICATE" in potential_key.id:
                raise UnsupportedCombinationError(
                    f"Combination failed due to key collision on {potential_key}. "
                    "Some keys already have _DUPLICATE appended to their ID. "
                    "Please report this issue if you need this functionality.",
                )

    for collection_name, collection in input2.collections.items():
        # TODO: Actually specify behavior in this case
        try:
            new_collection = result.collections[collection_name]
        except KeyError:
            result.collections[collection_name] = collection
            warnings.warn(
                f"One Interchange object has collection with name {collection_name} not "
                f"found in the other Interchange object, but it has now been added.",
            )
            continue

        for top_key, pot_key in collection.key_map.items():
            _tmp_pot_key = copy.deepcopy(pot_key)
            new_atom_indices = tuple(idx + atom_offset for idx in top_key.atom_indices)
            new_top_key = top_key.__class__(**top_key.model_dump())
            try:
                new_top_key.atom_indices = new_atom_indices  # type: ignore[misc]
            except (ValueError, AttributeError):
                assert len(new_atom_indices) == 1
                new_top_key.this_atom_index = new_atom_indices[0]  # type: ignore
            # If interchange was not created with SMIRNOFF, we need avoid merging potentials with same key
            if pot_key.associated_handler == "ExternalSource":
                _mult = 0
                while _tmp_pot_key in new_collection.potentials:
                    _tmp_pot_key.mult = _mult
                    _mult += 1

            if _tmp_pot_key in input1[collection_name].potentials:
                # if the potential key and parameters are identical, no action needed
                if (
                    collection.potentials[_tmp_pot_key] == input1[collection_name].potentials[_tmp_pot_key]
                ) and collection_name != "Constraints":
                    pass

                else:
                    logging.info(f"Key collision with different parameters, fixing. Key is {_tmp_pot_key}")
                    _tmp_pot_key.id += "_DUPLICATE"

            new_collection.key_map.update(
                {new_top_key: _tmp_pot_key},
            )

            new_collection.potentials.update(
                {_tmp_pot_key: collection.potentials[pot_key]},
            )

        # Ensure the charge cache is rebuilt
        if collection_name == "Electrostatics":
            new_collection._charges_cached = False  # type: ignore[attr-defined]
            new_collection._get_charges()  # type: ignore[attr-defined]

        result.collections[collection_name] = new_collection

    if result.positions is not None and input2.positions is not None:
        result.positions = numpy.vstack([result.positions, input2.positions])
    else:
        warnings.warn(
            "Setting positions to None because one or both objects added together were missing positions.",
            InterchangeCombinationWarning,
        )
        result.positions = None

    if not numpy.all(result.box == result.box):
        raise UnsupportedCombinationError(
            "Combination with unequal box vectors is not currently supported",
        )

    return result



import pytest
from openff.toolkit import Quantity

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer
from openff.interchange.drivers import get_openmm_energies


def test_combine_after_from_openmm_with_mainline_openmm_force_field(
    popc,
    sage,
):
    # amber/lipid17.xml is not shipped with OpenMM
    pytest.importorskip("openmmforcefields")

    import openmm.app
    import openmm.unit

    ligand = MoleculeWithConformer.from_smiles("c1ccccc1")
    ligand._conformers[0] += Quantity([3, 3, 3], "angstrom")
    topology = ligand.to_topology()
    topology.box_vectors = Quantity([6, 6, 6], "nanometer")

    popc_topology = popc.to_topology()
    popc_topology.box_vectors = Quantity([6, 6, 6], "nanometer")

    lipid17 = openmm.app.ForceField("amber/lipid17.xml")

    popc_system = lipid17.createSystem(
        topology=popc_topology.to_openmm(),
        nonbondedMethod=openmm.app.PME,
        constraints=openmm.app.HBonds,
    )

    imported = Interchange.from_openmm(
        system=popc_system,
        topology=popc_topology,
        positions=popc_topology.get_positions(),
    )

    ligand_interchange = sage.create_interchange(topology)

    for item in [imported, ligand_interchange]:
        item["vdW"].cutoff = Quantity(9, "angstrom")
        item["vdW"].switch_width = Quantity(1, "angstrom")
        item["Electrostatics"].cutoff = Quantity(9, "angstrom")

    combined = imported.combine(ligand_interchange)

    combined.to_openmm()

    energies = {
        "ligand": get_openmm_energies(ligand_interchange),
        "popc": get_openmm_energies(imported),
        "combined": get_openmm_energies(combined),
    }

    for key in ["Bond", "Angle", "Torsion"]:
        assert (energies["ligand"][key] + energies["popc"][key]).m == pytest.approx(
            energies["combined"][key].m,
        ), key

    # Non-bonded energies are not linear sums of component non-bonded energies, so
    # just make sure it's not a NaN (though a NaN should probably cause a crash ...)
    assert isinstance(energies["combined"]["Nonbonded"].m, float)



from copy import deepcopy

import numpy
import pytest
from openff.toolkit import Quantity
from openff.units import ensure_quantity
from openff.utilities import has_package, skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer
from openff.interchange._tests._openmm import get_14_scaling_factors

if has_package("openmm"):
    from openmm import (
        CMMotionRemover,
        HarmonicAngleForce,
        HarmonicBondForce,
        NonbondedForce,
        PeriodicTorsionForce,
    )


@skip_if_missing("openmm")
class TestOpenMM:
    def test_no_nonbonded_force(self, sage):
        """
        Ensure a SMIRNOFF-style force field can be exported to OpenMM even if no nonbonded handlers are present. For
        context, see https://github.com/openforcefield/openff-toolkit/issues/1102
        """

        del sage._parameter_handlers["Constraints"]
        del sage._parameter_handlers["NAGLCharges"]
        del sage._parameter_handlers["LibraryCharges"]
        del sage._parameter_handlers["Electrostatics"]
        del sage._parameter_handlers["vdW"]

        methane = MoleculeWithConformer.from_smiles("C")

        openmm_system = Interchange.from_smirnoff(sage, [methane]).to_openmm()

        for force in openmm_system.getForces():
            if isinstance(force, NonbondedForce):
                pytest.fail("A NonbondedForce was found in the OpenMM system.")
            elif isinstance(force, PeriodicTorsionForce):
                assert force.getNumTorsions() == 0
            elif isinstance(force, HarmonicBondForce):
                assert force.getNumBonds() == 4
            elif isinstance(force, HarmonicAngleForce):
                assert force.getNumAngles() == 6
            elif isinstance(force, CMMotionRemover):
                pass
            else:
                pytest.fail(f"Unexpected force found, type: {type(force)}")

    def test_14_scale_factors_missing_electrostatics(self, sage):
        # Ported from the toolkit after #1276
        topology = MoleculeWithConformer.from_smiles("CCCC").to_topology()

        ff_no_electrostatics = deepcopy(sage)
        ff_no_electrostatics.deregister_parameter_handler("Electrostatics")
        ff_no_electrostatics.deregister_parameter_handler("NAGLCharges")
        ff_no_electrostatics.deregister_parameter_handler("LibraryCharges")

        out = Interchange.from_smirnoff(
            ff_no_electrostatics,
            topology,
        ).to_openmm(combine_nonbonded_forces=True)

        numpy.testing.assert_almost_equal(
            actual=get_14_scaling_factors(out)[1],
            desired=ff_no_electrostatics["vdW"].scale14,
            decimal=8,
        )

    def test_14_scale_factors_missing_vdw(self, sage):
        # Ported from the toolkit after #1276
        topology = MoleculeWithConformer.from_smiles("CCCC").to_topology()

        ff_no_vdw = deepcopy(sage)
        ff_no_vdw.deregister_parameter_handler("vdW")

        out = Interchange.from_smirnoff(
            ff_no_vdw,
            topology,
        ).to_openmm(combine_nonbonded_forces=True)

        numpy.testing.assert_almost_equal(
            actual=get_14_scaling_factors(out)[0],
            desired=ff_no_vdw["Electrostatics"].scale14,
            decimal=8,
        )

    def test_to_pdb_box_vectors(self, sage):
        """Reproduce https://github.com/openforcefield/openff-interchange/issues/548."""
        from openmm.app import PDBFile

        topology = MoleculeWithConformer.from_smiles("CC").to_topology()
        topology.box_vectors = Quantity(
            10.0 * numpy.eye(3),
            "angstrom",
        )

        interchange = Interchange.from_smirnoff(sage, topology)

        interchange.to_pdb("temp.pdb")

        parsed_box_vectors = PDBFile("temp.pdb").topology.getPeriodicBoxVectors()

        numpy.testing.assert_allclose(
            topology.box_vectors.m_as("angstrom"),
            ensure_quantity(parsed_box_vectors, "openff").m_as("angstrom"),
        )


@skip_if_missing("openmm")
class TestOpenMMMissingHandlers:
    def test_missing_vdw_combine_energies(self, sage):
        from openff.interchange.drivers import get_openmm_energies

        topology = MoleculeWithConformer.from_smiles("CC").to_topology()

        ff_no_vdw = deepcopy(sage)
        ff_no_vdw.deregister_parameter_handler("vdW")

        out = Interchange.from_smirnoff(ff_no_vdw, topology)

        energy1 = get_openmm_energies(out, combine_nonbonded_forces=True).total_energy
        energy2 = get_openmm_energies(out, combine_nonbonded_forces=False).total_energy

        assert abs(energy2 - energy1) < Quantity(
            1e-6,
            "kilojoule_per_mole",
        )



"""Assorted utilities used in tests."""

import pathlib
import random
from importlib import resources

import numpy
import pytest
from openff.toolkit import (
    AmberToolsToolkitWrapper,
    ForceField,
    Molecule,
    OpenEyeToolkitWrapper,
    RDKitToolkitWrapper,
    Topology,
)
from openff.utilities import get_data_file_path
from openff.utilities.utilities import has_executable, has_package

from openff.interchange import Interchange
from openff.interchange.drivers.gromacs import _find_gromacs_executable

if has_package("openmm"):
    import openmm
    import openmm.unit

    kj_nm2_mol = openmm.unit.kilojoule_per_mole / openmm.unit.nanometer**2
    kj_rad2_mol = openmm.unit.kilojoule_per_mole / openmm.unit.radian**2

requires_ambertools = pytest.mark.skipif(
    not AmberToolsToolkitWrapper.is_available(),
    reason="Test requires AmberTools",
)
requires_rdkit = pytest.mark.skipif(
    not RDKitToolkitWrapper.is_available(),
    reason="Test requires RDKit",
)
requires_openeye = pytest.mark.skipif(
    not OpenEyeToolkitWrapper.is_available(),
    reason="Test requires OE toolkit",
)


_rng = numpy.random.default_rng(12345)


def get_test_file_path(test_file: str) -> pathlib.Path:
    """Given a filename in the collection of data files, return its full path."""
    test_dir_path = get_test_files_dir_path()
    test_file_path = test_dir_path / test_file

    if test_file_path.is_file():
        return test_file_path
    else:
        raise FileNotFoundError(
            f"could not file file {test_file} in path {test_file_path}",
        )


def get_test_files_dir_path(dirname: str | None = None) -> pathlib.Path:
    """Given a directory with a collection of test data files, return its full path."""
    dir_path = resources.files("openff.interchange._tests.data")

    if dirname:
        test_dir: pathlib.PosixPath = dir_path / dirname  # type: ignore[assignment]
    else:
        test_dir = dir_path  # type: ignore[assignment]

    if test_dir.is_dir():
        return test_dir
    else:
        raise NotADirectoryError(
            f"Provided directory {dirname} doesn't exist in {dir_path}",
        )


class MoleculeWithConformer(Molecule):
    """Thin wrapper around `Molecule` to produce an instance with a conformer in one call."""

    @classmethod
    def from_smiles(self, smiles, name="", *args, **kwargs):
        """Create from smiles and generate a single conformer."""
        molecule = super().from_smiles(smiles, *args, **kwargs)
        molecule.generate_conformers(n_conformers=1)
        molecule.name = name

        return molecule

    @classmethod
    def from_mapped_smiles(self, smiles, name="", **kwargs):
        """Create from smiles and generate a single conformer."""
        molecule = super().from_mapped_smiles(smiles, **kwargs)
        molecule.generate_conformers(n_conformers=1)
        molecule.name = name

        return molecule


def get_protein(name: str) -> Molecule:
    """Get a protein from openff/toolkit/data/proteins based on PDB name."""
    return Topology.from_pdb(
        get_data_file_path(
            relative_path=f"proteins/{name}.pdb",
            package_name="openff.toolkit",
        ),
    ).molecule(0)


HAS_GROMACS = _find_gromacs_executable() is not None
HAS_LAMMPS = has_package("lammps")
HAS_SANDER = has_executable("sander")

needs_gmx = pytest.mark.skipif(not HAS_GROMACS, reason="Needs GROMACS")
needs_not_gmx = pytest.mark.skipif(
    HAS_GROMACS,
    reason="Needs GROMACS to NOT be installed",
)
needs_lmp = pytest.mark.skipif(not HAS_LAMMPS, reason="Needs LAMMPS")
needs_not_lmp = pytest.mark.skipif(
    HAS_LAMMPS,
    reason="Needs LAMMPS to NOT be installed",
)
needs_sander = pytest.mark.skipif(not HAS_SANDER, reason="Needs sander")
needs_not_sander = pytest.mark.skipif(
    HAS_SANDER,
    reason="sander needs to NOT be installed",
)


def shuffle_topology(
    interchange: Interchange,
    force_field: ForceField,
    seed: int = 123455,
) -> Interchange:
    """Shuffle the atom order of the topology and return a new Interchange object."""
    assert interchange.positions is not None, "Written to also manage positions"

    rng = random.Random(seed)

    shuffled_topology = Topology()

    # can't easily shuffle a generator ...
    for molecule in interchange.topology.molecules:
        topology_indices = [interchange.topology.atom_index(atom) for atom in molecule.atoms]

        molecule._conformers = [
            interchange.positions[topology_indices, :].to("angstrom"),
        ]

        # so just randomize order at insertion time
        shuffled_topology._molecules.insert(
            rng.randint(0, len(shuffled_topology._molecules)),
            molecule,
        )

    shuffled_topology._invalidate_cached_properties()

    shuffled_topology.box_vectors = interchange.topology.box_vectors

    return force_field.create_interchange(shuffled_topology)



from openff.toolkit import Quantity
from pydantic import Field

from openff.interchange._annotations import _Quantity
from openff.interchange.pydantic import _BaseModel


class Person(_BaseModel):
    mass: _Quantity = Field()


class Roster(_BaseModel):
    people: dict[str, Person] = Field(dict())

    foo: _Quantity = Field()


class Model(_BaseModel):
    array: _Quantity


def test_simple_model_validation():
    bob = Person(mass="100.0 kilogram")

    assert Person.model_validate(bob) == bob
    assert Person.model_validate(bob.model_dump()) == bob

    assert Person.model_validate_json(bob.model_dump_json()) == bob


def test_simple_model_setter():
    bob = Person(mass="100.0 kilogram")

    bob.mass = "90.0 kilogram"

    assert bob.mass == Quantity("90.0 kilogram")


def test_model_with_array_quantity():
    model = Model(array=Quantity([1, 2, 3], "angstrom"))

    for test_model in [
        Model.model_validate(model),
        Model.model_validate(model.model_dump()),
        Model.model_validate_json(model.model_dump_json()),
    ]:
        assert all(test_model.array == model.array)


def test_nested_model():
    roster = Roster(
        people={"Bob": {"mass": "100.0 kilogram"}, "Alice": {"mass": "70.0 kilogram"}},
        foo="10.0 year",
    )

    assert Roster.model_validate(roster) == roster
    assert Roster.model_validate(roster.model_dump()) == roster

    assert Roster.model_validate_json(roster.model_dump_json()) == roster



"""Storing and processing results of energy evaluations."""

import warnings
from typing import Annotated

from openff.toolkit import Quantity
from pydantic import BeforeValidator, Field

from openff.interchange._annotations import _Quantity
from openff.interchange.exceptions import (
    EnergyError,
    IncompatibleTolerancesError,
    InvalidEnergyError,
)
from openff.interchange.pydantic import _BaseModel

_KNOWN_ENERGY_TERMS = [
    "Bond",
    "Angle",
    "Torsion",
    "RBTorsion",
    "Nonbonded",
    "vdW",
    "vdW 1-4",
    "Electrostatics",
    "Electrostatics 1-4",
]


def energies_validator(value: dict[str, Quantity | None]) -> dict[str, Quantity | None]:
    """Validate a dict of energies."""
    if not isinstance(value, dict):
        raise ValueError(f"wrong input type{type(value)}")

    for key, val in value.items():
        if key not in _KNOWN_ENERGY_TERMS:
            raise InvalidEnergyError(f"Energy type {key} not understood.")

        if val is None:
            continue

        if "openmm" in str(type(val)):
            from openff.units.openmm import from_openmm

            value[key] = from_openmm(val).to("kilojoule / mole")
            continue

        if isinstance(val, Quantity):
            value[key] = val.to("kilojoule / mole")

        else:
            raise InvalidEnergyError(f"Energy type {key} not understood.")

    # return sorted in a consistent order
    return {key: value[key] for key in _KNOWN_ENERGY_TERMS if key in value}


_EnergiesDict = Annotated[
    dict[str, _Quantity | None],
    BeforeValidator(energies_validator),
]


class EnergyReport(_BaseModel):
    """A lightweight class containing single-point energies as computed by energy tests."""

    # TODO: Should the default be None or 0.0 kj_mol?
    energies: _EnergiesDict = Field(
        {
            "Bond": None,
            "Angle": None,
            "Torsion": None,
            "vdW": None,
            "Electrostatics": None,
        },
    )

    @property
    def total_energy(self):
        """Return the total energy."""
        return self["total"]

    def __getitem__(self, item: str) -> Quantity | None:
        if type(item) is not str:
            raise LookupError(
                f"Only str arguments can be currently be used for lookups.\nFound item {item} of type {type(item)}",
            )
        if item in self.energies.keys():
            return self.energies[item]
        if item.lower() == "total":
            return sum(self.energies.values())  # type: ignore
        else:
            return None

    def update(self, new_energies: dict) -> None:
        """Update the energies in this report with new value(s)."""
        self.energies.update(energies_validator(new_energies))

    def compare(
        self,
        other: "EnergyReport",
        tolerances: dict[str, Quantity] | None = None,
    ):
        """
        Compare two energy reports.

        Parameters
        ----------
        other
            The other `EnergyReport` to compare energies against

        tolerances
            Per-key allowed differences in energies

        """
        default_tolerances = {
            "Bond": Quantity(1e-3, "kilojoule / mole"),
            "Angle": Quantity(1e-3, "kilojoule / mole"),
            "Torsion": Quantity(1e-3, "kilojoule / mole"),
            "vdW": Quantity(1e-3, "kilojoule / mole"),
            "Electrostatics": Quantity(1e-3, "kilojoule / mole"),
        }

        if tolerances:
            default_tolerances.update(tolerances)

        tolerances = default_tolerances

        # Ensure everything is in kJ/mol for safety of later comparison
        energy_differences = {key: diff.to("kilojoule / mole") for key, diff in self.diff(other).items()}

        if ("Nonbonded" in tolerances) != ("Nonbonded" in energy_differences):
            raise IncompatibleTolerancesError(
                "Mismatch between energy reports and tolerances with respect to whether nonbonded "
                "interactions are collapsed into a single value.",
            )

        errors = dict()

        for key, diff in energy_differences.items():
            if abs(energy_differences[key]) > tolerances[key]:
                errors[key] = diff

        if errors:
            raise EnergyError(errors)

    def diff(
        self,
        other: "EnergyReport",
    ) -> dict[str, Quantity]:
        """
        Return the per-key energy differences between these reports.

        Parameters
        ----------
        other
            The other `EnergyReport` to compare energies against

        Returns
        -------
        energy_differences
            Per-key energy differences

        """
        energy_differences: dict[str, Quantity] = dict()

        nonbondeds_processed = False

        for key in self.energies:
            if key in ("Bond", "Angle", "Torsion"):
                try:
                    energy_differences[key] = self[key] - other[key]  # type: ignore[operator]

                except KeyError:
                    continue

            if key in ("Nonbonded", "vdW", "Electrostatics"):
                if nonbondeds_processed:
                    continue

                if (self["vdW"] and other["vdW"]) is not None and (
                    self["Electrostatics"] and other["Electrostatics"]
                ) is not None:
                    for key in ("vdW", "Electrostatics"):
                        energy_differences[key] = self[key] - other[key]  # type: ignore[operator]
                        energy_differences[key] = self[key] - other[key]  # type: ignore[operator]

                        nonbondeds_processed = True

                        continue

                else:
                    energy_differences["Nonbonded"] = self._get_nonbonded_energy() - other._get_nonbonded_energy()

                    nonbondeds_processed = True

                    continue

        return energy_differences

    def __sub__(self, other: "EnergyReport") -> dict[str, Quantity]:
        diff = dict()
        for key in self.energies:
            if key not in other.energies:
                warnings.warn(f"Did not find key {key} in second report")
                continue
            diff[key]: Quantity = self.energies[key] - other.energies[key]  # type: ignore

        return diff

    def __str__(self) -> str:
        return (
            "Energies:\n\n"
            f"Bond:          \t\t{self['Bond']}\n"
            f"Angle:         \t\t{self['Angle']}\n"
            f"Torsion:       \t\t{self['Torsion']}\n"
            f"RBTorsion:     \t\t{self['RBTorsion']}\n"
            f"Nonbonded:     \t\t{self['Nonbonded']}\n"
            f"vdW:           \t\t{self['vdW']}\n"
            f"Electrostatics:\t\t{self['Electrostatics']}\n"
        )

    def _get_nonbonded_energy(self) -> Quantity:
        nonbonded_energy = Quantity(0.0, "kilojoule / mole")
        for key in ("Nonbonded", "vdW", "Electrostatics"):
            if key in self.energies is not None:
                nonbonded_energy += self.energies[key]

        return nonbonded_energy
